import os
from typing import Any

from amino.map import Map
from amino import Either
from amino.io import IO


class EnvVars:

    @property
    def vars(self):
        return Map(os.environ)

    def __contains__(self, name: str) -> bool:
        return self.contains(name)

    def contains(self, name: str) -> bool:
        return name in self.vars

    def __getitem__(self, name: str) -> Either[str, str]:
        return self.get(name)

    def get(self, name: str) -> Either[str, str]:
        return self.vars.lift(name).to_either(f'env var {name} is unset')

    def __setitem__(self, name: str, value: str) -> None:
        self.set(name, value)

    def set(self, name: str, value: str) -> None:
        os.environ[name] = str(value)

    def __delitem__(self, name: str) -> None:
        return self.delete(name)

    def delete(self, name: str) -> None:
        if name in os.environ:
            del os.environ[name]

env = EnvVars()


def set_env(name: str, value: Any) -> IO[None]:
    return IO.delay(env.set, name, str(value))


__all__ = ('EnvVars', 'env')



from typing import Any

from amino import Path, List, IO, do, Do, Nil
from amino.meta.gen import codegen, CodegenTask, codegen_write

here = Path(__file__).absolute().parent
template_path = here.parent.parent / 'templates' / 'state.py'


def state_task(
        tpe: str,
        path: str,
        tvar: List[str]=Nil,
        class_extra: str='',
        meta_extra: str='',
        ctor_extra: str='',
        extra_import: List[str]=Nil,
        extra: str='',
) -> CodegenTask:
    tpar = tvar.map(lambda a: f'{a}, ').mk_string('')
    tvars = tvar.map(lambda a: f'{a} = TypeVar(\'{a}\')').join_lines
    tpe_import = f'from {path} import {tpe}'
    subs = List(
        (r'\bF\[', f'{tpe}[{tpar}'),
        (r'\bF\b', tpe),
        ('StateT\[', f'{tpe}State[{tpar}'),
        ('StateT', f'{tpe}State'),
        ('{tpar}', tpar),
        ('{tvar}', tvars),
        ('{f_import}', tpe_import),
        ('{class_extra}', class_extra),
        ('{meta_extra}', meta_extra),
        ('{ctor_extra}', ctor_extra),
        ('{extra_import}', extra_import.join_lines),
        ('{extra}', extra),
        ('StateBase', 'StateT'),
    )
    return CodegenTask(template_path, subs, Nil)


def generate_state(write: bool, name: str, *a: Any, **kw: Any) -> IO[str]:
    task = state_task(*a, **kw)
    outpath = here.parent / 'state' / f'{name}.py'
    return codegen_write(task, outpath) if write else codegen(task)


@do(IO[None])
def generate_states() -> Do:
    yield generate_state(True, 'maybe', 'Maybe', 'amino.maybe')
    yield generate_state(True, 'eval', 'Eval', 'amino.eval')
    yield generate_state(True, 'io', 'IO', 'amino.io')
    yield generate_state(True, 'id', 'Id', 'amino.id')
    yield generate_state(True, 'either', 'Either', 'amino.either', tvar=List('E'))


__all__ = ('generate_state', 'generate_states',)



import abc
import json
from typing import TypeVar, Generic, Type

from amino.tc.base import TypeClass
from amino import Either, L, _, Map, do, Do, Lists
from amino.json.data import JsonError, Json, JsonObject, JsonScalar, tpe_key, JsonArray
from amino.logging import module_log
from amino.util.tpe import qualname

A = TypeVar('A')
log = module_log()


class Encoder(Generic[A], TypeClass):

    @abc.abstractmethod
    def encode(self, a: A) -> Either[JsonError, Json]:
        ...


@do(Either[JsonError, Json])
def encode_json(data: A) -> Do:
    enc = yield Encoder.e_for(data).lmap(L(JsonError)(data, _))
    yield enc.encode(data)


@do(Either[JsonError, str])
def dump_json(data: A) -> Do:
    js = yield encode_json(data)
    return json.dumps(js.native)


def json_type(tpe: Type) -> Json:
    mod = '__builtins__' if tpe.__module__ == 'builtins' else tpe.__module__
    names = Lists.split(qualname(tpe), '.').map(JsonScalar)
    return JsonObject(Map(module=JsonScalar(mod), names=JsonArray(names)))


def json_object_with_type(fields: Map[str, JsonObject], tpe: Type[A]) -> Json:
    return JsonObject(fields.cat((tpe_key, json_type(tpe))))


__all__ = ('Encoder', 'encode_json', 'dump_json')



from typing import TypeVar, Callable, Generic, Tuple, cast

from amino.tc.base import F, ImplicitsMeta, Implicits
from amino import Either, List, options, Do, Boolean
from amino.state import State
from amino.do import do
from amino.dat import ADT, ADTMeta
from amino.case import Case, CaseRec, Term, RecStep
from amino.logging import module_log
from amino.compute.data import CResult, CFatal, Thunk, CSuccess, CError, eval_thunk
from amino.compute.exception import ComputeException

log = module_log()
A = TypeVar('A')
B = TypeVar('B')
C = TypeVar('C')
S = TypeVar('S')
debug = options.io_debug.exists


class ComputeMeta(ImplicitsMeta, ADTMeta):
    pass


class Compute(
        Generic[S, A],
        Implicits,
        ADT['Compute[S, A]'],
        implicits=True,
        imp_mod='amino.compute.tc',
        metaclass=ComputeMeta,
):

    def eval(self) -> State[S, CResult[A]]:
        return eval_compute(self)

    def run(self, vim: S) -> Tuple[S, CResult[A]]:
        try:
            return self.eval().run(vim).value
        except ComputeException as e:
            return vim, CFatal(e)

    def run_s(self, vim: S) -> S:
        return self.run(vim)[0]

    def run_a(self, vim: S) -> CResult[A]:
        return self.run(vim)[1]

    result = run_a

    def either(self, vim: S) -> Either[ComputeException, A]:
        return self.result(vim).to_either

    def unsafe_run(self, vim: S) -> Tuple[S, A]:
        result_vim, result = self.run(vim)
        return result_vim, result.to_either.get_or_raise()

    def unsafe(self, vim: S) -> A:
        return self.either(vim).get_or_raise()


class ComputePure(Generic[S, A], Compute[S, A]):

    def __init__(self, value: A) -> None:
        self.value = value


class ComputeSuspend(Generic[S, A], Compute[S, A]):

    @staticmethod
    def cons(thunk: State[S, Compute[S, A]]) -> 'ComputeSuspend[S, A]':
        return ComputeSuspend(Thunk.cons(thunk))

    def __init__(self, thunk: Thunk[S, Compute[S, A]]) -> None:
        self.thunk = thunk


class ComputeBind(Generic[S, A, B], Compute[S, B]):

    def __init__(
            self,
            thunk: Thunk[S, Compute[S, A]],
            kleisli: Callable[[A], Compute[S, B]],
    ) -> None:
        self.thunk = thunk
        self.kleisli = kleisli


class ComputeError(Generic[S, A], Compute[S, A]):

    def __init__(self, error: str) -> None:
        self.error = error


class ComputeFatal(Generic[S, A], Compute[S, A]):

    def __init__(self, exception: Exception) -> None:
        self.exception = exception


class ComputeRecover(Generic[S, A], Compute[S, A]):

    def __init__(
            self,
            io: Compute[S, A],
            recover: Callable[[CResult[A]], Compute[S, A]],
            recoverable: Callable[[CResult[A]], Boolean],
    ) -> None:
        self.io = io
        self.recover = recover
        self.recoverable = recoverable


class lift_n_result(Case[CResult[A], Compute[S, A]], alg=CResult):

    def n_success(self, result: CSuccess[A]) -> Compute[S, A]:
        return ComputePure(result.value)

    def n_error(self, result: CError[A]) -> Compute[S, A]:
        return ComputeError(result.error)

    def n_fatal(self, result: CFatal[A]) -> Compute[S, A]:
        return ComputeFatal(result.exception)


@do(State[S, Compute[S, Either[List[str], A]]])
def execute_compute_recover(io: ComputeRecover[S, A]) -> Do:
    def execute_wrapped(vim: S) -> Tuple[S, Either[List[str], A]]:
        updated_vim, result = io.io.run(vim)
        return updated_vim, (
            io.recover(result)
            if io.recoverable(result) else
            lift_n_result.match(result)
        )
    updated_vim, result = yield State.inspect(execute_wrapped)
    yield State.reset(updated_vim, result)


class flat_map_compute(Generic[S, A, B], Case[Callable[[A], Compute[S, B]], Compute[S, B]], alg=Compute):

    def __init__(self, f: Callable[[A], Compute[S, B]]) -> None:
        self.f = f

    def compute_pure(self, io: ComputePure[S, A]) -> Compute[S, B]:
        thunk = State.delay(self.f, io.value)
        return ComputeSuspend.cons(thunk)

    def compute_suspend(self, io: ComputeSuspend[S, A]) -> Compute[S, B]:
        return ComputeBind(io.thunk, self.f)

    def compute_bind(self, io: ComputeBind[S, C, A]) -> Compute[S, B]:
        return ComputeSuspend.cons(State.pure(ComputeBind(io.thunk, lambda a: io.kleisli(a).flat_map(self.f))))

    def compute_error(self, io: ComputeError[S, A]) -> Compute[S, B]:
        return cast(Compute[S, B], io)

    def compute_fatal(self, io: ComputeFatal[S, A]) -> Compute[S, B]:
        return cast(Compute[S, B], io)

    def compute_recover(self, io: ComputeRecover[S, A]) -> Compute[S, B]:
        return ComputeBind(Thunk.cons(execute_compute_recover(io)), self.f)


EvalRes = RecStep[Compute[S, A], Tuple[S, CResult[A]]]


class eval_step(Generic[S, A], CaseRec[Compute[S, A], CResult[A]], alg=Compute):

    def __init__(self, vim: S) -> None:
        self.vim = vim

    def compute_pure(self, io: ComputePure[S, A]) -> EvalRes:
        return Term((self.vim, CSuccess(io.value)))

    def compute_suspend(self, io: ComputeSuspend[S, A]) -> EvalRes:
        vim, next = eval_thunk(self.vim, io.thunk)
        return eval_step(vim)(next)

    def compute_bind(self, io: ComputeBind[S, B, A]) -> EvalRes:
        vim, next = eval_thunk(self.vim, io.thunk)
        return eval_step(vim)(next.flat_map(io.kleisli))

    def compute_error(self, io: ComputeError[S, A]) -> EvalRes:
        return Term((self.vim, CError(io.error)))

    def compute_fatal(self, io: ComputeFatal[S, A]) -> EvalRes:
        return Term((self.vim, CFatal(io.exception)))

    def compute_recover(self, io: ComputeRecover[S, A]) -> Compute[S, B]:
        '''calls `map` to shift the recover execution to flat_map_compute
        '''
        return eval_step(self.vim)(io.map(lambda a: a))


@do(State[S, A])
def eval_compute(io: Compute[S, A]) -> Do:
    vim = yield State.get()
    updated_vim, result = eval_step(vim)(io).eval()
    yield State.set(updated_vim)
    return result


__all__ = ('Compute',)



import abc
from typing import Callable, Iterable, TypeVar, Generic, List

import amino  # NOQA
from amino.tc.apply import Apply
from amino.func import I
from amino.tc.base import tc_prop
from amino.tc.apply_n import ApplyN

F = TypeVar('F')
G = TypeVar('G')
A = TypeVar('A')
B = TypeVar('B')


class FlatMap(Apply, ApplyN):

    def apply_n_funcs(self) -> List[str]:
        return super().apply_n_funcs() + ['flat_map', 'product']

    def ap(self, fa: F, ff: F):
        f = lambda f: self.map(fa, f)
        return self.flat_map(ff, f)

    @abc.abstractmethod
    def flat_map(self, fa: F, f: Callable[[A], G]) -> G:
        ...

    def __floordiv__(self, fa, f):
        return self.flat_map(fa, f)

    def product(self, fa: F, fb: F) -> F:
        f = lambda a: self.map(fb, lambda b: (a, b))
        return self.flat_map(fa, f)

    __and__ = product

    def product_n(self, num: int, fa: F, *fs: Iterable[F]):
        from amino.list import List
        if len(fs) != num:
            msg = 'passed {} args to {}.product{}'
            name = self.__class__.__name__
            raise TypeError(msg.format(len(fs), name, num))
        def add(a, b):
            return self.flat_map(a, lambda a: self.map(b, lambda b: a + (b,)))
        init = self.map(fa, lambda a: (a,))
        return List.wrap(fs).fold_left(init)(add)

    def flat_pair(self, fa: F, f: Callable[[A], 'amino.maybe.Maybe[B]']) -> F:
        cb = lambda a: f(a).map(lambda b: (a, b))
        return self.flat_map(fa, cb)

    def flat_replace(self, fa: F, fb: F) -> F:
        cb = lambda a: fb
        return self.flat_map(fa, cb)

    def and_then(self, fa: F, fb: F) -> F:
        cb = lambda a: fb
        return self.flat_map(fa, cb)

    @tc_prop
    def join(self, fa: F) -> F:
        return self.flat_map(fa, I)

__all__ = ('FlatMap',)



import os
import sys
import subprocess


def compile(path: str) -> None:
    if subprocess.run(['coconut', '--target', '3.6', '--quiet', '--strict', path]).returncode != 0:
        print('failed to compile coco files')
        sys.exit(1)


def compile_coco_projects() -> None:
    return
    if 'COCO_DIRS' in os.environ and 'AMINO_COMPILE_COCO' in os.environ:
        for dir in os.environ.get('COCO_DIRS', '').split(':'):
            compile(dir)

__all__ = ('compile_coco_projects',)



import inspect
from types import FrameType
from typing import Any, Tuple

from amino import List, Lists, Path, Either, _, Try, Right, Maybe, L, Nil
from amino.do import do, Do
from amino.func import tailrec


def cframe() -> FrameType:
    return inspect.currentframe()


@do(Either[Exception, str])
def file_line(file: str, lineno: int) -> Do:
    path = yield Try(Path, file)
    text = yield Try(path.read_text)
    yield Lists.lines(text).lift(lineno).to_either('frame_data: invalid line number in frame')


def frame_data(frame: FrameType) -> Tuple[str, str]:
    file = inspect.getsourcefile(frame.f_code)
    code = file_line(file, frame.f_lineno - 1).value_or(lambda err: f'invalid file entry `{file}`: {err}')
    return file, code


def traceback_entry(frame: FrameType, file: str, code: str) -> List[str]:
    line = frame.f_lineno
    fun = frame.f_code.co_name
    clean = code.strip()
    return List(f'  File "{file}", line {line}, in {fun}', f'    {clean}')


def frame_traceback_entry(frame: FrameType) -> List[str]:
    file, code = frame_data(frame)
    return traceback_entry(frame, file, code)


default_internal_packages = List('amino')


def internal_frame(frame: FrameType, pkgs: List[str]=None) -> bool:
    pkg = frame.f_globals.get('__package__')
    name = frame.f_code.co_filename or ''
    return (
        (not isinstance(pkg, str)) or
        pkg == '' or
        (pkgs or default_internal_packages).exists(pkg.startswith) or
        not name.endswith('.py')
    )


def non_internal_frame(frame: FrameType, pkgs: List[str]=None) -> bool:
    return not internal_frame(frame, pkgs)


def tb_callsite(tb: List[FrameType], pkgs: List[str]=None) -> Either[str, FrameType]:
    return tb.find(L(non_internal_frame)(_, pkgs)).to_either('no non-internal frame found')


def frame_callsite(frame: FrameType, pkgs: List[str]=None) -> FrameType:
    @tailrec
    def loop(f: FrameType) -> FrameType:
        return (True, (f.f_back,)) if not (f is None or f.f_back is None) and internal_frame(f, pkgs) else (False, f)
    return loop(frame)


def callsite_traceback_entry(frame: FrameType, pkgs: List[str]=None) -> List[str]:
    return frame_traceback_entry(frame_callsite(frame, pkgs))


def callsite_info(frame: FrameType, pkgs: List[str]=None) -> List[str]:
    return Maybe.optional(frame) / L(callsite_traceback_entry)(_, pkgs) | List('  <no callsite info>')


def callsite_source(frame: FrameType, pkgs: List[str]=None) -> str:
    return Maybe.optional(frame) / (lambda f: frame_data(frame_callsite(f, pkgs))) / _[1] | '<no source>'


__all__ = ('cframe', 'callsite_info', 'callsite_source', 'tb_callsite', 'non_internal_frame', 'internal_frame')



from types import FunctionType
from typing import Type, Tuple, Any

from amino import List, Map


def type_tail(a: Any) -> str:
    return str(a).split('.')[-1]


def typename(a: Any) -> str:
    return getattr(a, __name__, type_tail(a))


def synth_method(name: str, params: List[Tuple[str, Type]], statements: List[str], _globals: dict) -> FunctionType:
    id = f'synth_{name}__'
    params_s = params.map2(lambda n, t: f'{n}: {typename(t)}').join_comma
    param_globals = Map(params.map2(lambda n, t: (typename(t), t)))
    globs = Map(_globals) ** param_globals
    code = f'''\
def {name}(self, {params_s}) -> None:
{statements.indent(4).join_lines}
globals()['{id}'] = {name}
    '''
    exec(code, globs)
    return globs.pop(id)


def synth_init(params: List[Tuple[str, Type]], _globals: dict) -> FunctionType:
    assign = params.map2(lambda k, v: f'self.{k} = {k}')
    return synth_method('__init__', params, assign, _globals)


__all__ = ('synth_method', 'synth_init')



from functools import wraps, partial
from inspect import getfullargspec
from typing import Callable, Union, Any, TypeVar, Tuple, Generic, cast

A = TypeVar('A')
B = TypeVar('B')
C = TypeVar('C')


def curried(func: Callable[..., B]) -> Callable[[A], Callable[..., Union[Callable, B]]]:
    @wraps(func)
    def _curried(*args: Any, **kwargs: Any) -> Union[B, Callable[..., Union[Callable, B]]]:
        f = func
        count = 0
        while isinstance(f, partial):
            if f.args:
                count += len(f.args)
            f = f.func
        spec = getfullargspec(f)
        if count == len(spec.args) - len(args):
            return func(*args, **kwargs)
        else:
            return curried(partial(func, *args, **kwargs))
    return _curried


class Identity:

    def __init__(self) -> None:
        self.__name__ = 'identity'

    def __call__(self, a: A) -> A:
        return a

    def __str__(self) -> str:
        return '(a => a)'

I = Identity()


class Val(Generic[A]):

    def __init__(self, value: A) -> None:
        self.value = value
        self.__name__ = self.__class__.__name__

    def __call__(self) -> A:
        return self.value

    def __str__(self) -> str:
        return '{}({})'.format(self.__class__.__name__, self.value)


class ReplaceVal(Generic[A], Val[A]):

    def __call__(self, *a: Any, **kw: Any) -> A:
        return super().__call__()


def flip(a: A, b: B) -> Tuple[B, A]:
    return b, a


CallByName = Union[Any, Callable[[], Any]]


def call_by_name(b: Union[A, Callable[[], A]]) -> A:
    return b() if callable(b) else cast(A, b)


def is_not_none(a: Any) -> bool:
    return a is not None


def tupled2(f: Callable[[A, B], C]) -> Callable[[Tuple[A, B]], C]:
    def wrap(a: Tuple[A, B]) -> C:
        return f(a[0], a[1])
    return wrap


TailrecResult = Tuple[bool, Union[A, tuple]]


class tailrec:
    __slots__ = 'func',

    def __init__(self, func: Callable[..., Tuple[bool, Union[A, tuple]]]) -> None:
        self.func = func

    def __call__(self, *a: Any) -> A:
        args: Union[A, tuple] = a
        while True:
            cont, args = self.func(*cast(tuple, args))
            if not cont:
                break
        return cast(A, args)


tco = tailrec


class mtailrec:
    __slots__ = 'func',


def const(a: A) -> Callable[[B], A]:
    return lambda b: a


__all__ = ('curried', 'I', 'flip', 'call_by_name', 'Val', 'ReplaceVal', 'is_not_none', 'tupled2', 'tailrec', 'mtailrec',
           'const', 'TailrecResult',)



from typing import TypeVar, Callable, Tuple, Type
import itertools

from amino import Maybe, LazyList, _, __
from amino.list import List
from amino.func import curried, call_by_name, CallByName
from amino.tc.monad import Monad
from amino.tc.base import ImplicitInstances, tc_prop, F
from amino.lazy import lazy
from amino.tc.traverse import Traverse
from amino.tc.foldable import Foldable, FoldableABC
from amino.tc.zip import Zip
from amino.instances.list import ListZip
from amino.tc.optional import Optional

A = TypeVar('A')
B = TypeVar('B')


class LazyListInstances(ImplicitInstances):

    @lazy
    def _instances(self):
        from amino.map import Map
        return Map(
            {
                Monad: LazyListMonad(),
                Traverse: LazyListTraverse(),
                Foldable: LazyListFoldable(),
                Zip: LazyListZip(),
            }
        )


class LazyListMonad(Monad):

    def pure(self, a: A):
        return LazyList([], List(a))

    def map(self, fa: LazyList[A], f: Callable[[A], B]) -> LazyList[B]:
        return fa.copy(lambda a: map(f, a), __.map(f))

    def flat_map(self, fa: LazyList[A], f: Callable[[A], LazyList[B]]) -> LazyList[B]:
        a, b = itertools.tee(fa.source)
        fa.source = a
        strict_m = fa.strict.map(f)
        lazy_m = map(f, b)
        mapped = itertools.chain(strict_m, lazy_m)
        source = itertools.chain.from_iterable(mapped)
        return LazyList(source, List(), fa._chunk_size)


class LazyListTraverse(Traverse):

    def traverse(self, fa: LazyList[A], f: Callable, tpe: type):
        return fa.drain.traverse(f, tpe) / LazyList


FoldableABC.register(LazyList)


class LazyListFoldable(Foldable):

    @tc_prop
    def with_index(self, fa: LazyList[A]) -> List[Tuple[int, A]]:
        return LazyList(enumerate(fa.source), fa.strict, fa._chunk_size)

    def filter(self, fa: LazyList[A], f: Callable[[A], bool]) -> LazyList[A]:
        return fa.copy(lambda l: filter(f, l), __.filter(f))

    @curried
    def fold_left(self, fa: LazyList[A], z: B, f: Callable[[B, A], B]) -> B:
        return Foldable.fatal(List).fold_left(fa.drain, z, f)

    def find(self, fa: List[A], f: Callable[[A], bool]) -> Maybe[A]:
        return fa.strict.find(f).o(fa._drain_find(f))

    def find_map(self, fa: LazyList[A], f: Callable[[A], Maybe[B]]) -> Maybe[B]:
        return fa.map(f).find(_.is_just)

    def index_where(self, fa: LazyList[A], f: Callable[[A], bool]
                    ) -> Maybe[int]:
        return fa.strict.index_where(f) | (
            fa._drain_find(f) / (lambda a: len(fa.strict) - 1))

    def find_map_optional(self, fa: List[A], tpe: Type[F], f: Callable[[A], F[B]], msg: CallByName=None) -> F[B]:
        a = fa.map(f).find(_.present)
        return a | (lambda: Optional.fatal(tpe).absent(call_by_name(msg)))


class LazyListZip(Zip):

    def _zip(self, fs):
        return zip(*map(_.source, fs))

    def zip(self, fa: LazyList[A], fb: LazyList[B], *fs) -> LazyList:
        fss = (fa, fb) + fs
        maxlen = max(map(lambda a: len(a.strict), fss))
        for f in fss:
            f._fetch(maxlen - 1)
        stricts = map(_.strict, fss)
        strict = ListZip().zip(*stricts)
        return LazyList(self._zip(fss), init=strict)

__all__ = ('LazyListInstances',)



from __future__ import annotations
import abc
import inspect
from types import SimpleNamespace, FunctionType
from typing import TypeVar, Type, Any, Generic, cast, Tuple, get_type_hints

from amino import Map, Lists, List, Nil, _, Maybe, Just, L, Try
from amino.util.string import ToStr
from amino.func import Val
from amino.lazy import lazy
from amino.algebra import AlgebraMeta, Algebra
from amino.tc.base import ImplicitsMeta
from amino.util.tpe import qualified_type

A = TypeVar('A')


class KeepField:
    pass


class Field(ToStr):

    def __init__(self, name: str, tpe: Type) -> None:
        self.name = name
        self.tpe = tpe

    @property
    def qualified_type(self) -> None:
        return Try(qualified_type, self.tpe).get_or_strict(self.tpe)

    def _arg_desc(self) -> List[str]:
        return List(self.name, self.qualified_type)

    @property
    def param_str(self) -> str:
        return f'''{self.name}: '{self.qualified_type}'=Dat.Keep'''


Sub = TypeVar('Sub', bound='Dat')


class FieldMutator(Generic[Sub]):

    def __init__(self, name: str, target: 'Dat[Sub]') -> None:
        self.name = name
        self.target = target

    @abc.abstractmethod
    def __call__(self, value: Any) -> Sub:
        ...


class FieldSetter(Generic[Sub], FieldMutator[Sub]):

    def __call__(self, value: Any) -> Sub:
        return self.target.copy(**{self.name: value})


class FieldModifier(Generic[Sub], FieldMutator[Sub]):

    def __call__(self, f) -> Sub:
        return self.target.copy(**{self.name: f(getattr(self.target, self.name))})


class FieldAppender(Generic[Sub], FieldMutator[Sub]):

    def __call__(self, value) -> Sub:
        return self.target.mod(self.name)(_ + value)


class FieldAppender1(Generic[Sub], FieldMutator[Sub]):

    def __call__(self, value) -> Sub:
        return self.target.mod(self.name)(_ + List(value))


class FieldProxy(Generic[Sub], FieldMutator[Sub]):

    def __init__(self, target: Sub, tpe: Type[FieldMutator[Sub]]) -> None:
        self.target = target
        self.tpe = tpe

    def __getattr__(self, name: str) -> FieldMutator[Sub]:
        return self(name)

    def __call__(self, name: str) -> FieldMutator[Sub]:
        return self.tpe(name, self.target)


def init_fields(init: FunctionType, globalns: dict) -> List[Field]:
    spec = inspect.getfullargspec(init)
    args = Lists.wrap(spec.args).tail | Nil
    types = Map(get_type_hints(init))
    def field(name: str) -> Field:
        tpe = types.lift(name) | Val(Any)
        return Field(name, tpe)
    return args / field


class DatMeta(abc.ABCMeta):

    def __new__(cls: type, name: str, bases: tuple, namespace: SimpleNamespace, **kw) -> type:
        mod = inspect.currentframe()
        caller = mod.f_back
        globalns = caller.f_globals
        fs = Map(namespace).lift('__init__') / (lambda a: init_fields(a, globalns)) | Nil
        inst = super().__new__(cls, name, bases, namespace, **kw)
        if not (fs.empty and hasattr(inst, '_dat__fields_value')):
            inst._dat__fields_value = fs
        return inst

    @property
    def _dat__fields(self) -> List[Field]:
        return self._dat__fields_value

    @property
    def _field_count_min(self) -> int:
        return len(self._dat__fields)

    @property
    def _field_count_max(self) -> Maybe[int]:
        return Just(len(self._dat__fields))


def to_str(a: Any) -> str:
    return (
        a.__name__
        if isinstance(a, FunctionType) and hasattr(a, '__name__') else
        str(a)
    )


# TODO typecheck ctor args in development
class Dat(Generic[Sub], ToStr, metaclass=DatMeta):
    Keep = KeepField()

    @property
    def _dat__fields(self) -> List[Field]:
        return type(self)._dat__fields

    @lazy
    def _dat__names(self) -> List[str]:
        return self._dat__fields.map(_.name)

    @lazy
    def _dat__values(self) -> List[Any]:
        return (
            self._dat__fields
            .traverse(lambda a: Maybe.getattr(self, a.name), Maybe)
            .get_or_fail(lambda: f'corrupt `Dat`: {type(self)}')
        )

    @lazy
    def _dat__items(self) -> List[Tuple[str, Any]]:
        return self._dat__names.zip(self._dat__values)

    @property
    def to_map(self) -> Map[str, Any]:
        return Map(self._dat__items)

    def copy(self, **kw: Any) -> Sub:
        updates = Map(kw)
        def update(f: Field) -> Any:
            return updates.lift(f.name) | (lambda: getattr(self, f.name))
        updated = self._dat__fields / update
        return cast(Dat, type(self)(*updated))

    def typed_copy(self, **kw: Any) -> Sub:
        updates = Map(kw)
        def update(f: Field) -> Any:
            return updates.lift(f.name).filter_type(f.tpe) | (lambda: getattr(self, f.name))
        updated = self._dat__fields / update
        return cast(Dat, type(self)(*updated))

    @property
    def append(self) -> FieldProxy[Sub]:
        return FieldProxy(self, FieldAppender)

    @property
    def append1(self) -> FieldProxy[Sub]:
        return FieldProxy(self, FieldAppender1)

    @property
    def mod(self) -> FieldProxy[Sub]:
        return FieldProxy(self, FieldModifier)

    @property
    def set(self) -> FieldProxy[Sub]:
        return FieldProxy(self, FieldSetter)

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, type(self)) and self._dat__values == other._dat__values

    def __iter__(self) -> iter:
        return iter(self._dat__values)

    def __hash__(self) -> int:
        return hash(self._dat__values)

    def _lens_setattr(self, name, value):
        return self.set(name)(value)

    def _arg_desc(self) -> List[str]:
        return self._dat__values / to_str


class DatImplicitsMeta(ImplicitsMeta, DatMeta):
    pass


class ADTMeta(DatMeta, AlgebraMeta):
    pass


class ADT(Algebra, Dat[Sub], metaclass=ADTMeta, algebra_base=True):
    pass


class ADTImplicitsMeta(ImplicitsMeta, ADTMeta):
    pass


__all__ = ('Dat', 'ADT', 'DatImplicitsMeta', 'ADTImplicitsMeta',)



from types import GeneratorType
from typing import TypeVar, Callable, Any, Generator, Type
import functools

from amino.tc.monad import Monad

A = TypeVar('A')
B = TypeVar('B')
Do = Generator


# NOTE ostensibly, this cannot be tailrecced without separating strictly evaluated monadic composition from lazy ones.
# itr.gi_frame.f_lasti is the instruction pointer and could be used to detect laziness.
# NOTE due to the nature of generators, a do with a lazily evaluated monad cannot be executed twice.
# NOTE Lists don't work properly because the generator will be consumed by the first element
def do_impl(tpe: type, f: Callable[..., Generator[A, B, A]]) -> Callable[..., A]:
    @functools.wraps(f)
    def do_loop(*a: Any, **kw: Any) -> A:
        m = Monad.fatal(tpe)
        itr = f(*a, **kw)
        if not isinstance(itr, GeneratorType):
            raise Exception(f'function `{f.__qualname__}` decorated with `do` does not produce a generator')
        @functools.wraps(f)
        def loop(val: B) -> A:
            try:
                return m.flat_map(itr.send(val), loop)
            except StopIteration as e:
                return m.pure(val if e.value is None else e.value)
        return m.flat_map(m.pure(None), loop)
    return do_loop


def do(tpe: Type[A]) -> Callable[[Callable[..., Generator[A, B, A]]], Callable[..., A]]:
    def deco(f: Callable[..., Generator[A, B, A]]) -> Callable[..., A]:
        f.tpe = tpe  # type: ignore
        f.__do = None  # type: ignore
        f.__do_original = f  # type: ignore
        return functools.wraps(f)(do_impl)(tpe, f)
    return deco

tdo = do

__all__ = ('do', 'tdo', 'Do')



from typing import Union, Any, TypeVar, Type, Callable, Tuple

import amino
from amino import maybe
from amino.either import Right, Left, Either
from amino.func import call_by_name

A = TypeVar('A')
B = TypeVar('B')


class Boolean:

    def __init__(self, value: Union['Boolean', bool]) -> None:
        self.value = bool(value)

    @staticmethod
    def wrap(value):
        return Boolean(value)

    @staticmethod
    def issubclass(value: Type[A], tpe: Type[B]) -> 'Boolean':
        return Boolean(isinstance(value, type) and issubclass(value, tpe))

    @staticmethod
    def isinstance(value: A, tpe: Union[Type[Any], Tuple[Type[Any], ...]]) -> 'Boolean':
        return Boolean(isinstance(value, tpe))

    @staticmethod
    def is_a(tpe: Union[Type[Any], Tuple[Type[Any], ...]]) -> Callable[[Any], 'Boolean']:
        return lambda a: Boolean.isinstance(a, tpe)

    @staticmethod
    def is_a_class(tpe: Type[A]) -> Callable[[Any], 'Boolean']:
        return lambda a: Boolean.issubclass(a, tpe)

    def maybe(self, value):
        return maybe.Maybe.optional(value) if self else maybe.Empty()

    def flat_maybe(self, value: 'Maybe'):  # type: ignore
        return value if self else maybe.Empty()

    def maybe_call(self, f, *a, **kw):
        return maybe.Just(f(*a, **kw)) if self else maybe.Empty()

    def m(self, v):
        return maybe.Maybe.optional(call_by_name(v)) if self else maybe.Empty()

    def flat_maybe_call(self, f, *a, **kw):
        return f(*a, **kw) if self else maybe.Empty()

    def flat_m(self, v):
        return call_by_name(v) if self else maybe.Empty()

    def either(self, l, r):
        return self.either_call(l, lambda: r)

    def either_call(self, l, r):
        return Right(r()) if self else Left(l)

    def flat_either_call(self, l, r):
        return r() if self else Left(l)

    def e(self, f: A, t: B) -> Either[A, B]:
        return Right(call_by_name(t)) if self else Left(call_by_name(f))

    def flat_e(self, l, r):
        return call_by_name(r) if self else Left(call_by_name(l))

    def l(self, v: A) -> 'amino.List[A]':
        return self.m(v) / amino.List | amino.Nil

    def cata(self, t: A, f: A) -> A:
        return t if self.value else f

    def cata_call(self, t, f):
        return t() if self.value else f()

    def c(self, t, f):
        return call_by_name(t) if self.value else call_by_name(f)

    def __bool__(self):
        return self.value

    def __str__(self):
        return '⊤' if self.value else '⊥'

    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__, self.value)

    def __eq__(self, other):
        return (
            self.value == other
            if isinstance(other, bool) else
            self.value == other.value
            if isinstance(other, Boolean) else
            False
        )

    def __and__(self, other: Any) -> 'Boolean':
        return Boolean(self and other)

    def __or__(self, other: Any) -> 'Boolean':
        return Boolean(self or other)

    def __invert__(self) -> 'Boolean':
        return Boolean(not self.value)

    def __xor__(self, other: Any) -> 'Boolean':
        return Boolean(bool(self.value ^ bool(other)))

    def __rxor__(self, other: Any) -> 'Boolean':
        return Boolean(bool(self.value ^ bool(other)))

    def __hash__(self) -> int:
        return hash(self.value)

    @property
    def no(self):
        return Boolean(not self.value)

    @property
    def json_repr(self):
        return self.value

    @property
    def to_int(self) -> int:
        return 1 if self else 0


true = Boolean(True)
false = Boolean(False)

__all__ = ('Boolean', 'true', 'false')



from typing import Any, Callable, TypeVar, Type
from types import ModuleType

from amino.do import do, Do
from amino import Maybe, Lists, L, _, Either, List


@do(Either[str, List[Any]])
def objects_from_module(mod: ModuleType, pred: Callable[[Any], bool]) -> Do:
    all = yield Maybe.getattr(mod, '__all__').to_either(f'module `{mod.__name__}` does not define `__all__`')
    return (
        Lists.wrap(all)
        .flat_map(L(Maybe.getattr)(mod, _))
        .filter(pred)
    )


@do(Either[str, List[Any]])
def object_from_module(mod: ModuleType, pred: Callable[[Any], bool], desc: str) -> Do:
    all = yield objects_from_module(mod, pred)
    yield all.head.to_either(f'no {desc} in `{mod.__name__}.__all__`')


AS = TypeVar('AS')
A = TypeVar('A', bound=AS)


def cls_from_module(mod: ModuleType, tpe: Type[AS]) -> Either[str, Type[A]]:
    pred = lambda a: isinstance(a, type) and issubclass(a, tpe)
    return object_from_module(mod, pred, f'subclass of `{tpe}`')


def instances_from_module(mod: ModuleType, tpe: Type[A]) -> Either[str, List[A]]:
    pred = lambda a: isinstance(a, tpe)
    return objects_from_module(mod, pred)


def instance_from_module(mod: ModuleType, tpe: Type[A]) -> Either[str, A]:
    pred = lambda a: isinstance(a, tpe)
    return object_from_module(mod, pred, f'instance of `{tpe.__name__}`')


__all__ = ('object_from_module', 'cls_from_module', 'instance_from_module', 'instances_from_module',)



from typing import TypeVar, Dict, Generic, Tuple, Callable

from toolz import dicttoolz

from amino import Maybe, may, Just, Nothing
from amino.data.list import List, Lists
from amino.boolean import Boolean
from amino.tc.base import Implicits, ImplicitsMeta

A = TypeVar('A')
B = TypeVar('B', covariant=True)
C = TypeVar('C')
D = TypeVar('D', covariant=True)


class Map(Generic[A, B], Dict[A, B], Implicits, metaclass=ImplicitsMeta, implicits=True):

    @staticmethod
    def wrap(d: Dict[A, B]) -> 'Map[A, B]':
        return Map(d)

    def lift(self, key: str) -> Maybe[B]:
        return Just(Dict.get(self, key)) if key in self else Nothing

    def lift_item(self, key):
        return self.lift(key) / (lambda a: (key, a))

    def lift_all(self, *keys):
        def append(zm, k):
            return zm // (lambda z: self.lift(k) / z.cat)
        return List.wrap(keys).fold_left(Just(List()))(append)

    def get_all_map(self, *keys):
        def append(zm, k):
            return zm // (lambda z: self.get_item(k) / z.cat)
        return List.wrap(keys).fold_left(Just(Map()))(append)

    def lift_or_else(self, key, default: Callable[[], C]):
        return self.lift(key).get_or_else(default)

    def set_if_missing(self, key: A, default: Callable[[], B]):
        if key in self:
            return self
        else:
            return self + (key, default())

    def __str__(self):
        return str(dict(self))

    def cat(self, item: Tuple[A, B]):
        return Map(dicttoolz.assoc(self, *item))

    __add__ = cat

    def merge(self, other: 'Map[A, B]'):
        return Map(dicttoolz.merge(self, other))

    __pow__ = merge

    def __sub__(self, key: A):
        return Map(dicttoolz.dissoc(self, key))

    def find(self, f: Callable[[B], bool]) -> Maybe[Tuple[A, B]]:
        return self.to_list.find(lambda item: f(item[1]))

    def find_key(self, f: Callable[[A], bool]) -> Maybe[Tuple[A, B]]:
        return self.to_list.find(lambda item: f(item[0]))

    def valfilter(self, f: Callable[[B], bool]) -> 'Map[A, B]':
        return Map(dicttoolz.valfilter(f, self))

    def keyfilter(self, f: Callable[[A], bool]) -> 'Map[A, B]':
        return Map(dicttoolz.keyfilter(f, self))

    def filter(self, f: Callable[[Tuple[A, B]], bool]) -> 'Map[A, B]':
        return Map(dicttoolz.itemfilter(f, self))

    def valmap(self, f: Callable[[B], C]) -> 'Map[A, C]':
        return Map(dicttoolz.valmap(f, dict(self)))

    def keymap(self, f: Callable[[A], C]) -> 'Map[C, B]':
        return Map(dicttoolz.keymap(f, dict(self)))

    def map(self, f: Callable[[A, B], Tuple[C, D]]) -> 'Map[C, D]':
        wrap = lambda a: f(*a)
        return Map(dicttoolz.itemmap(wrap, self))

    def flat_map(
            self,
            f: Callable[[A, B], Maybe[Tuple[C, D]]]
    ) -> 'Map[C, D]':
        filtered = List.wrap([f(a, b) for a, b in self.items()])\
            .join
        return Map(filtered)

    def bimap(self, fa: Callable[[A], C], fb: Callable[[B], D]) -> 'Map[C, D]':
        return self.map(lambda a, b: (fa(a), fb(b)))

    def map2(self, f: Callable[[A, B], C]) -> List[C]:
        return List.wrap([f(a, b) for a, b in self.items()])

    def flat_map2(self, f: Callable[[A, B], List[C]]) -> List[C]:
        return self.to_list.flat_map2(f)

    @property
    def to_list(self):
        return List.wrap(self.items())

    @property
    def head(self):
        return self.to_list.head

    @property
    def keys_view(self):
        return Dict.keys(self)

    @property
    def values_view(self):
        return Dict.values(self)

    @property
    def k(self):
        return List(*Dict.keys(self))

    @property
    def v(self) -> List[B]:
        return Lists.wrap(Dict.values(self))

    @property
    def is_empty(self):
        return self.k.is_empty

    empty = is_empty

    def at(self, *keys):
        return self.keyfilter(lambda a: a in keys)

    def values_at(self, *keys):
        return List.wrap(keys) // self.lift

    def has_key(self, name):
        return Boolean(name in self)

    contains = has_key

    def add_maybe(self, key: A, value: Maybe[B]):
        return value / (lambda a: self + (key, a)) | self

    def insert_if_absent(self, key: A, value: Callable[[], B]) -> 'Map[A, B]':
        return self if key in self else self + (key, value())


__all__ = ('Map',)



from functools import singledispatch
import typing
from typing import Callable, Any, Dict, TypeVar, Type

from amino.util.string import snake_case
from amino.algebra import Algebra

A = TypeVar('A')
B = TypeVar('B')
R = TypeVar('R')
Alg = TypeVar('Alg', bound=Algebra)


def dispatch(obj: B, tpes: typing.List[A], prefix: str, default: Callable[[A], R]=None) -> Callable[[A], R]:
    def error(o: A) -> None:
        msg = 'no dispatcher defined for {} on {} {}'
        raise TypeError(msg.format(o, obj.__class__.__name__, prefix))
    @singledispatch
    def main(o: A, *a: Any, **kw: Any) -> R:
        if default is None:
            error(o)
        else:
            return default(o, *a, **kw)
    for tpe in tpes:
        fun = getattr(obj, '{}{}'.format(prefix, snake_case(tpe.__name__)), None)
        if fun is None:
            error(tpe)
        main.register(tpe)(fun)
    return main


def dispatch_alg(obj: B, alg: Type[Alg], prefix: str='', default: Callable[[Alg], R]=None) -> Callable[[Alg], R]:
    return dispatch(obj, alg.__algebra_variants__, prefix, default)


def dispatch_with(rules: Dict[type, Callable], default: Callable=None):
    @singledispatch
    def main(o, *a, **kw):
        if default is None:
            msg = 'no dispatcher defined for {} {} ({})'
            raise TypeError(msg.format(type(o), o, rules))
        else:
            default(o, *a, **kw)
    for tpe, fun in rules.items():
        main.register(tpe)(fun)
    return main


__all__ = ('dispatch', 'dispatch_alg', 'dispatch_with')



import re
from typing import Tuple

from amino import Path, IO, do, Do, Dat, List


class CodegenTask(Dat['CodegenTask']):

    @staticmethod
    def cons(
            template: Path,
            subs: List[Tuple[str, str]],
            imports: List[str],
    ) -> 'CodegenTask':
        return CodegenTask(
            template,
            subs,
            imports,
        )

    def __init__(
            self,
            template: Path,
            subs: List[Tuple[str, str]],
            imports: List[str],
    ) -> None:
        self.template = template
        self.subs = subs
        self.imports = imports


@do(IO[str])
def codegen(task: CodegenTask) -> Do:
    template_path = yield IO.delay(task.template.absolute)
    template_code = yield IO.delay(template_path.read_text)
    replaced = task.subs.fold_left(template_code)(lambda z, a: re.sub(a[0], a[1], z))
    return task.imports.cat(replaced).join_lines


@do(IO[str])
def codegen_write(task: CodegenTask, outpath: Path) -> Do:
    code = yield codegen(task)
    yield IO.delay(outpath.write_text, code)
    return code


__all__ = ('CodegenTask', 'codegen', 'codegen_write',)



from typing import TypeVar, Callable

from amino.tc.flat_map import FlatMap
from amino.tc.applicative import Applicative
from amino.data.list import List

F = TypeVar('F')
A = TypeVar('A')
B = TypeVar('B')


class Monad(FlatMap, Applicative):

    def map(self, fa: F, f: Callable[[A], B]) -> F:
        return self.flat_map(fa, lambda a: self.pure(f(a)))

    def eff(self, fa: F, tpe: type=None):
        from amino.eff import Eff
        tpes = List() if tpe is None else List(tpe)
        return Eff(fa, tpes, 1)

    def effs(self, fa: F, *args):
        from amino.eff import Eff
        types = List.wrap(args)
        c = lambda a, b: Eff(fa, a, depth=b)
        with_depth = lambda d, t: c(t, d)
        types_only = lambda: c(types, depth=len(types))
        def try_depth(h, t):
            return with_depth(int(h), t) if isinstance(h, int) else types_only()
        return types.detach_head.map2(try_depth) | types_only

__all__ = ('Monad',)



import abc
from typing import TypeVar, Type, Generic

from amino.tc.base import TypeClass
from amino import Either, Map, Right, Lists, Do, _, L, Left, Dat, List, Maybe
from amino.do import do
from amino.json.data import JsonError, JsonObject, JsonArray, JsonScalar, Json, JsonAbsent, JsonNull, tpe_key
from amino.json.parse import parse_json
from amino.case import Case
from amino.logging import module_log

A = TypeVar('A')
log = module_log()


class TypeInfo(Dat['TypeInfo']):

    def __init__(self, module: str, names: List[str]) -> None:
        self.module = module
        self.names = names


class Decoder(Generic[A], TypeClass):

    @abc.abstractmethod
    def decode(self, tpe: Type[A], data: Json) -> Either[JsonError, A]:
        ...


@do(Either[str, A])
def decode_json_object(data: dict) -> Do:
    m = Map(data)
    tpe_s = yield m.lift(tpe_key).to_either(f'no `{tpe_key}` attr in json object {m}')
    tpe = yield Either.import_path(tpe_s)
    dec = yield Decoder.e(tpe)
    yield dec.decode(tpe, m)


@do(Either[str, type])
def builtin_type(type_info: TypeInfo) -> Do:
    name = yield type_info.names.head.to_either(f'no name specified for builtin type')
    yield (
        Right(type(None))
        if name == 'NoneType' else
        Maybe.optional(__builtins__.get(name)).to_either(f'no builtin type named `{name}`')
    )


@do(Either[str, type])
def select_type(type_info: TypeInfo) -> Do:
    mod = yield Either.import_module(type_info.module)
    yield type_info.names.fold_m(Right(mod))(Either.getattr)


def decode_type_info(json: Json) -> Either[JsonError, TypeInfo]:
    return decode_json_type_json(json, TypeInfo)


@do(Either[JsonError, type])
def type_info(json: JsonObject) -> Do:
    jtpe = yield json.data.lift(tpe_key).to_either(json.error(f'no `{tpe_key}` field in json object'))
    type_info = yield decode_type_info(jtpe)
    yield (
        builtin_type(type_info)
        if type_info.module == '__builtins__' else
        select_type(type_info)
    ).lmap(json.error)


class decode(Generic[A], Case[Json, Either[JsonError, A]], alg=Json):

    @do(Either[JsonError, A])
    def decode_json_object(self, json: JsonObject) -> Do:
        tpe = yield type_info(json)
        dec = yield Decoder.e(tpe).lmap(L(JsonError)(json, _))
        yield dec.decode(tpe, json)

    def decode_json_array(self, json: JsonArray) -> Either[JsonError, A]:
        return Lists.wrap(json.data).traverse(self, Either)

    def decode_json_scalar(self, json: JsonScalar) -> Either[JsonError, A]:
        return Right(json.data)

    def decode_json_null(self, json: JsonNull) -> Either[JsonError, A]:
        return Right(None)

    def decode_json_absent(self, json: JsonAbsent) -> Either[JsonError, A]:
        return Left(JsonError(json.data), 'json absent')


@do(Either[JsonError, A])
def decode_json(data: str) -> Do:
    json = yield parse_json(data)
    yield decode.match(json)


@do(Either[JsonError, A])
def decode_json_type_json(json: Json, tpe: Type[A]) -> Do:
    dec = yield Decoder.e(tpe).lmap(L(JsonError)(json, _))
    yield dec.decode(tpe, json)


@do(Either[JsonError, A])
def decode_json_type(data: str, tpe: Type[A]) -> Do:
    json = yield parse_json(data)
    yield decode_json_type_json(json, tpe)


__all__ = ('Decoder', 'decode_json_object', 'decode_json', 'decode', 'decode_json_type', 'decode_json_type_json')



import abc
from typing import TypeVar

from amino.tc.base import TypeClass

A = TypeVar('A')


class Monoid(TypeClass):

    @abc.abstractproperty
    def empty(self) -> A:
        ...

    @abc.abstractmethod
    def combine(self, a: A, b: A) -> A:
        ...

__all__ = ('Monoid',)



from typing import Iterable, Any, TypeVar, _GenericAlias

from amino import do, Either, Do, Maybe, Right, Left, Lists
from amino.logging import module_log

A = TypeVar('A')
log = module_log()


def normalize_generic_type(tpe: type) -> type:
    return (
        tpe.__origin__
        if isinstance(tpe, _GenericAlias) else
        tpe
    )


def qualified_type(tpe: type) -> str:
    ntpe = normalize_generic_type(tpe)
    return ntpe.__name__ if ntpe.__module__ == 'builtins' else f'{ntpe.__module__}.{ntpe.__name__}'


def qualified_name(inst: Any) -> str:
    return inst.__name__ if inst.__module__ == 'builtins' else f'{inst.__module__}.{inst.__name__}'


@do(Either[str, type])
def type_arg(tpe: type, index: int) -> Do:
    def error() -> str:
        return f'{tpe} has no type args'
    raw = yield Maybe.getattr(tpe, '__args__').to_either_f(error)
    types = yield Right(Lists.wrap(raw)) if isinstance(raw, Iterable) else Left(error())
    yield types.lift(index).to_either_f(lambda: f'{tpe} has less than {index + 1} args')


def first_type_arg(tpe: type) -> Either[str, type]:
    return type_arg(tpe, 0)


def qualname(a: Any) -> str:
    return (
        a.__qualname__
        if hasattr(a, '__qualname__') else
        a.__origin__.__qualname__
        if hasattr(a, '__origin__') else
        a.__qualname__
    )


def is_subclass(tpe: Any, base: type) -> bool:
    def error():
        raise TypeError(f'argument `{tpe}` to `is_subclass({base})` is neither TypeVar nor type')
    return (
        False
        if isinstance(tpe, TypeVar) else
        issubclass(normalize_generic_type(tpe), base)
        if isinstance(tpe, (type, _GenericAlias)) else
        error()
    )


__all__ = ('qualified_type', 'type_arg', 'first_type_arg', 'qualified_name', 'qualname', 'normalize_generic_type',
           'is_subclass',)



import collections
from typing import Type, TypeVar, Collection, Mapping, Callable, Tuple, _SpecialForm
from numbers import Number
from uuid import UUID

from amino.json.decoder import (Decoder, decode, decode_json_type_json, type_info, TypeInfo, select_type,
                                decode_type_info)
from amino import (Maybe, Either, List, Lists, Left, Boolean, Try, Map, Right, Nothing, Just, Path, do, Do, _, L, ADT,
                   Dat)
from amino.json.data import JsonError, Json, JsonObject
from amino.dat import Field
from amino.logging import module_log

log = module_log()
A = TypeVar('A')
B = TypeVar('B')
Sub = TypeVar('Sub', bound=Dat)


class StringDecoder(Decoder, tpe=str):

    def decode(self, tpe: Type[str], data: Json) -> Either[JsonError, str]:
        return data.scalar.e(f'invalid type for `str`: {data}', data.data)


class NumberDecoder(Decoder, sub=Number):

    def decode(self, tpe: Type[int], data: Json) -> Either[JsonError, int]:
        return data.scalar.e(f'invalid type for `int`: {data}', data.data)


class MapDecoder(Decoder, sub=Mapping):

    def decode(self, tpe: Type[Mapping], data: Json) -> Either[JsonError, Map[str, A]]:
        def dec() -> Either[JsonError, Map[str, A]]:
            return Map(data.data).traverse(decode.match, Either)
        return data.object.c(dec, lambda: Left(f'invalid type for `Map`: {data}'))


class ListDecoder(Decoder, sub=Collection):

    def decode(self, tpe: Type[Collection], data: Json) -> Either[JsonError, List[A]]:
        def dec() -> Either[JsonError, List[A]]:
            return Lists.wrap(data.data).traverse(decode.match, Either)
        return data.array.c(dec, lambda: Left(f'invalid type for `List`: {data}'))


def maybe_from_object(data: JsonObject, inner: Maybe[Type[A]]) -> Either[JsonError, Maybe[A]]:
    return (
        decode.match(data) / Just
        if data.has_type else
        inner.traverse(lambda a: decode_json_type_json(data, a), Either)
    )


class MaybeDecoder(Decoder, tpe=Maybe):

    def decode(self, tpe: Type[Maybe], data: Json) -> Either[JsonError, Maybe[A]]:
        inner = Lists.wrap(tpe.__args__).head
        return (
            Right(Nothing)
            if data.absent else
            maybe_from_object(data, inner)
            if data.object else
            decode.match(data) / Just
            if data.array else
            Right(Nothing)
            if data.null else
            inner.cata(
                lambda a: decode_json_type_json(data, a) / Just,
                lambda: data.scalar.e(f'invalid type for `Maybe`: {data}', Maybe.check(data.data))
            )
        )


@do(Either[JsonError, Either[A, B]])
def either_from_object(data: JsonObject, ltype: Type[A], rtype: Type[B]) -> Do:
    value = yield data.field('value')
    decoded = yield decode.match(value)
    tpe = yield type_info(data)
    yield Right(tpe(decoded))


class EitherDecoder(Decoder, tpe=Either):

    @do(Either[JsonError, Either[A, B]])
    def decode(self, tpe: Type[Either], data: Json) -> Do:
        err = JsonError(data, f'too few types in Either')
        ltype, rtype = yield Lists.wrap(tpe.__args__).lift_all(0, 1).to_either(err)
        yield (
            either_from_object(data, ltype, rtype)
            if data.object else
            Left(JsonError(data, f'invalid type for `Either`'))
        )


class BooleanDecoder(Decoder, tpe=Boolean):

    def decode(self, tpe: Type[Boolean], data: Json) -> Either[JsonError, Boolean]:
        return data.scalar.e(f'invalid type for `Boolean`: {data}', Boolean(data.data))


class UUIDDecoder(Decoder, tpe=UUID):

    def decode(self, tpe: Type[UUID], data: Json) -> Either[JsonError, UUID]:
        return data.scalar.flat_e(f'invalid type for `UUID`: {data}', Try(UUID, data.data))


class PathDecoder(Decoder, tpe=Path):

    def decode(self, tpe: Type[Path], data: Json) -> Either[JsonError, Path]:
        return data.scalar.flat_e(lambda: JsonError(data, f'invalid type for `Path`'), Try(Path, data.data))


def decode_instance(data: Json, desc: str) -> Either[JsonError, A]:
    @do(Either[JsonError, A])
    def run(data: Json) -> Do:
        type_info = yield decode_json_type_json(data, TypeInfo)
        yield select_type(type_info)
    return (
        run(data)
        if data.object else
        Left(JsonError(data, f'invalid type for `{desc}`'))
    )


class CallableDecoder(Decoder, tpe=collections.abc.Callable):

    def decode(self, tpe: Type[Callable], data: Json) -> Either[JsonError, Callable]:
        return decode_instance(data, 'Callable')


class TupleDecoder(Decoder, tpe=tuple):

    @do(Either[JsonError, Tuple])
    def decode(self, tpe: Type[Tuple], data: Json) -> Do:
        t_data = yield data.field('data')
        a_data = yield t_data.as_array
        data = yield decode.match(a_data)
        yield Try(tuple, data)


@do(Either[JsonError, type])
def decode_type(data: Json) -> Do:
    info = yield decode_type_info(data)
    yield (
        Right(type(None))
        if info.module in ['builtins', '__builtins__'] and info.names == ['NoneType']
        else decode_instance(data, 'type')
    )


class TypeDecoder(Decoder, tpe=type):

    @do(Either[JsonError, Type])
    def decode(self, tpe: Type[Type], data: Json) -> Do:
        yield decode_type(data)


class TTypeDecoder(Decoder, tpe=Type):

    @do(Either[JsonError, Type])
    def decode(self, tpe: Type[Type], data: Json) -> Do:
        yield decode_type(data)


@do(Either[JsonError, A])
def decode_field_type(tpe: type, data: Json) -> Do:
    dec = yield Decoder.e(tpe).lmap(L(JsonError)(data, _))
    yield dec.decode(tpe, data)


def decode_field(data: Json) -> Do:
    @do(Either[JsonError, A])
    def decode_field(field: Field) -> Do:
        value = yield data.field(field.name)
        yield (
            Left(JsonError(data, f'invalid type for json: {field}'))
            if isinstance(field.tpe, _SpecialForm) else
            decode.match(value)
            if isinstance(field.tpe, TypeVar) else
            decode_field_type(field.tpe, value)
        )
    return decode_field


def decode_dat(tpe: Type[A], data: Json) -> Either[JsonError, A]:
    return tpe._dat__fields.traverse(decode_field(data), Either).map(lambda a: tpe(*a))


class DatDecoder(Decoder, tpe=Dat):

    def decode(self, tpe: Type[Sub], data: Json) -> Either[JsonError, Sub]:
        return decode_dat(tpe, data)


class ADTDecoder(Decoder, tpe=ADT):

    @do(Either[JsonError, Sub])
    def decode(self, tpe: Type[Sub], data: JsonObject) -> Do:
        sub_type = yield type_info(data)
        yield decode_dat(sub_type, data)


__all__ = ('MaybeDecoder', 'StringDecoder', 'NumberDecoder', 'ListDecoder', 'BooleanDecoder', 'MapDecoder',
           'PathDecoder', 'decode_instance', 'CallableDecoder', 'TupleDecoder', 'decode_type', 'TypeDecoder',
           'TTypeDecoder', 'ADTDecoder',)



from typing import List, TypeVar, Generic, Type, Dict, Callable, Optional

R = TypeVar('R')
A = TypeVar('A')
B = TypeVar('B')
C = TypeVar('C')


class Bound:

    def __init__(self, cls: str) -> None:
        self.cls = cls

    def __repr__(self) -> str:
        return f'Bound({self.cls})'


class Kind:
    pass


class Nullary(Kind):

    def __init__(self, name: str) -> None:
        self.name = name

    def __repr__(self) -> str:
        return self.name


class Wildcard(Kind):

    def __repr__(self) -> str:
        return '_'


class HK(Kind):

    def __init__(self, name: str, params: List[Kind]) -> None:
        self.name = name
        self.params = params

    def __repr__(self) -> str:
        params = ', '.join(map(str, self.params))
        return f'{self.name}[{params}]'


class Par:
    pass


class Mono(Par):

    def __init__(self, tpe: type) -> None:
        self.tpe = tpe

    def __repr__(self) -> str:
        return self.tpe


class Poly(Par):

    def __init__(self, kind: Kind, bounds: List[Bound]=[]) -> None:
        self.kind = kind
        self.bounds = bounds

    def __repr__(self) -> str:
        bounds = ''.join((f' <: {b}' for b in self.bounds))
        return f'{self.kind}{bounds}'


class Signature:

    def __init__(self, input: List[Par], output: Par) -> None:
        self.input = input
        self.output = output

    def __repr__(self) -> str:
        input = ', '.join(map(str, self.input))
        return f'({input}) => {self.output}'


class ClassFunction:
    pass


class ClassFunction1(Generic[R, A], ClassFunction):

    def __init__(self, name: str, signature: Signature, par: Type[A], ret: Type[R]) -> None:
        self.name = name
        self.signature = signature
        self.par = par
        self.ret = ret


class ClassFunction2(Generic[R, A, B], ClassFunction):

    def __init__(self, name: str, signature: Signature, par1: Type[A], par2: Type[B], ret: Type[R]) -> None:
        self.name = name
        self.signature = signature
        self.par1 = par1
        self.par2 = par2
        self.ret = ret
        self.cls: Optional[str] = None

    def set_class(self, cls: 'Class') -> None:
        self.cls = cls.name
        self.resolver = cons_resolver(cls, self.signature)

    def __call__(self, a: A) -> Callable[[B], R]:
        inst = find_instance(self.cls, type(a))


class Class:

    @staticmethod
    def cons(name: str, params: List[Par], functions: List[ClassFunction]) -> 'Class':
        cls = Class(name, params, [f for f in functions])
        for f in functions:
            f.set_class(cls)
        return cls


    def __init__(self, name: str, params: List[Par], functions: List[ClassFunction]) -> None:
        self.name = name
        self.params = params
        self.functions = functions


class Instance(Generic[A]):

    def __init__(self, cls: Class, tpe: Type[A], functions: Dict[str, ClassFunction]) -> None:
        self.cls = cls
        self.tpe = tpe
        self.functions = functions


class Instances:

    def __init__(self) -> None:
        self.registry: Dict[str, Dict[type, Instance]] = dict()


instances = Instances()


def register(cls: str, inst: Instance[A]) -> None:
    instances.registry.setdefault(cls, dict())
    print(inst.tpe)
    instances.registry[cls][inst.tpe] = inst


def instance(cls: Class, tpe: A, **functions: ClassFunction) -> None:
    inst = Instance(cls, tpe, functions)
    register(cls.name, inst)


def find_instance(name: str, tpe: Type[A]) -> None:
    return instances.registry[name][tpe]


class HK1(Generic[A]):
    pass


def cons_resolver(cls: Class, signature: Signature) -> Callable:
    print(cls.params)
    print(signature)
    def resolver(a) -> None:
        return
    return resolver


__all__ = ('Bound', 'Kind', 'Nullary', 'HK', 'Par', 'Mono', 'Poly', 'Signature', 'ClassFunction', 'Class', 'Instance',
           'Wildcard',)



import abc
from typing import TypeVar, Union, Generic
from numbers import Number

from amino import Map, List, _, Either, Right, Left, Boolean
from amino.util.string import ToStr
from amino.algebra import Algebra

A = TypeVar('A')
tpe_key = '__type__'


class JsonError(ToStr):

    def __init__(self, data: Union['Json', str], error: Union[str, Exception]) -> None:
        self.data = data
        self.error = error

    def _arg_desc(self) -> List[str]:
        return List(self.data, str(self.error))

    @property
    def exception(self) -> Exception:
        return self.error if isinstance(self.error, Exception) else Exception(self.error)


class Json(Algebra[A]):

    def __init__(self, data: A) -> None:
        self.data = data

    @abc.abstractproperty
    def native(self) -> Union[dict, list, str, Number, None]:
        ...

    @abc.abstractmethod
    def field(self, key: str) -> Either[JsonError, 'Json']:
        ...

    def _arg_desc(self) -> List[str]:
        return List(str(self.data))

    @property
    def scalar(self) -> Boolean:
        return Boolean.isinstance(self, JsonScalar)

    @property
    def array(self) -> Boolean:
        return Boolean.isinstance(self, JsonArray)

    @property
    def object(self) -> Boolean:
        return Boolean.isinstance(self, JsonObject)

    @property
    def absent(self) -> Boolean:
        return Boolean.isinstance(self, JsonAbsent)

    @property
    def null(self) -> Boolean:
        return Boolean.isinstance(self, JsonNull)

    def error(self, msg: Union[str, Exception]) -> JsonError:
        return JsonError(self.data, msg)

    @property
    def as_scalar(self) -> Either[JsonError, 'JsonScalar']:
        return Right(self) if self.scalar else Left(self.error('not a scalar'))

    @property
    def as_array(self) -> Either[JsonError, 'JsonArray']:
        return Right(self) if self.array else Left(self.error('not an array'))

    @property
    def as_object(self) -> Either[JsonError, 'JsonObject']:
        return Right(self) if self.object else Left(self.error('not an object'))


class JsonObject(Json[Map[str, Json]]):

    @property
    def native(self) -> Union[dict, list, str, Number, None]:
        return self.data.valmap(_.native)

    def field(self, key: str) -> Either[JsonError, Json]:
        return Right(self.data.lift(key) | JsonAbsent(self.error(f'no field `{key}`')))

    @property
    def has_type(self) -> Boolean:
        return self.data.contains(tpe_key)


class JsonArray(Json[List[Json]]):

    @property
    def native(self) -> Union[dict, list, str, Number, None]:
        return self.data.map(_.native)

    def field(self, key: str) -> Either[JsonError, Json]:
        return Left(self.error('JsonArray has no fields'))


class JsonScalar(Json[Union[str, Number]]):

    @property
    def native(self) -> Union[dict, list, str, Number]:
        return self.data

    def field(self, key: str) -> Either[JsonError, Json]:
        return Left(self.error('JsonScalar has no fields'))


class JsonNull(Json[None]):

    @staticmethod
    def cons() -> 'Json[None]':
        return JsonNull(None)

    @property
    def native(self) -> None:
        None

    def field(self, key: str) -> Either[JsonError, Json]:
        return Left(self.error('JsonNull has no fields'))


class JsonAbsent(Json[JsonError]):

    @property
    def native(self) -> Union[dict, list, str, Number, None]:
        return None

    def field(self, key: str) -> Either[JsonError, Json]:
        return Right(self)


__all__ = ('JsonError', 'Json', 'JsonObject', 'JsonArray', 'JsonScalar', 'JsonAbsent')



from typing import TypeVar, Callable, Tuple

from amino.data.list import List, Lists, Nil
from amino import Eval, Maybe, Nothing, Just, __, Either

A = TypeVar('A')
B = TypeVar('B')


def replace_one(fa: List[A], pred: Callable[[A], bool], new: A) -> Eval[Maybe[List[A]]]:
    def loop(head: A, tail: List[A]) -> Eval[Maybe[List[A]]]:
        def flatten(memla: Maybe[Eval[Maybe[List[A]]]]) -> Eval[Maybe[List[A]]]:
            return (
                memla
                .map(lambda emla: emla.map(lambda mla: mla.map(lambda la: la.cons(head)))) |
                (lambda: Eval.now(Nothing))
            )
        return (
            Eval.now(Just(tail.cons(new)))
            if pred(head) else
            Eval.later(tail.detach_head.map2, loop) // flatten
        )
    return fa.detach_head.map2(loop) | Eval.now(Nothing)


def split_by_status_zipped(data: List[Tuple[A, bool]]) -> Tuple[List[A], List[A]]:
    def decide(z: Tuple[List[A], List[A]], item: Tuple[A, bool]) -> Tuple[List[A], List[A]]:
        l, r = z
        a, status = item
        return (
            (l.cat(a), r)
            if status else
            (l, r.cat(a))
        )
    return data.fold_left((Nil, Nil))(decide)


def split_by_status(data: List[A], statuses: List[bool]) -> Tuple[List[A], List[A]]:
    return split_by_status_zipped(data.zip(statuses))


def split_either_list(data: List[Either[A, B]]) -> Tuple[List[A], List[B]]:
    def split(z: Tuple[List[A], List[B]], a: Either[A, B]) -> Tuple[List[A], List[B]]:
        ls, rs = z
        return a.cata((lambda e: (ls.cat(e), rs)), (lambda v: (ls, rs.cat(v))))
    return data.fold_left((Nil, Nil))(split)


__all__ = ('List', 'Lists', 'Nil', 'split_by_status', 'split_by_status_zipped', 'split_either_list',)



from amino.json.decoder import decode_json
from amino.json.encoder import encode_json, dump_json
import amino.json.encoders
import amino.json.decoders

__all__ = ('decode_json', 'encode_json', 'dump_json')



import abc
import time
import inspect
import typing
import traceback
from pathlib import Path
from threading import Thread
from typing import Callable, TypeVar, Generic, Any, Union, Tuple, Awaitable, Optional

from amino import Either, Right, Left, Maybe, List, __, Just, Lists, L, Nothing, options, _
from amino.eval import Eval
from amino.tc.base import Implicits, ImplicitsMeta
from amino.logging import log
from amino.util.fun import lambda_str, format_funcall
from amino.util.string import ToStr
from amino.do import do, Do
from amino.func import tailrec
from amino.util.trace import callsite_info, frame_callsite, callsite_source, frame_traceback_entry, non_internal_frame
from amino.util.exception import format_exception_error, format_exception
from amino.options import io_debug

A = TypeVar('A')
B = TypeVar('B')
C = TypeVar('C')


class IOExceptionBase(Exception, abc.ABC):

    @abc.abstractproperty
    def desc(self) -> str:
        ...

    @abc.abstractproperty
    def internal_packages(self) -> Maybe[List[str]]:
        ...

    def __init__(self, f, stack, cause, frame=None) -> None:
        self.f = f
        self.stack = List.wrap(stack)
        self.cause = cause
        self.frame = frame

    @property
    def lines(self) -> List[str]:
        return self.full_trace_lines if io_debug else self.truncated_trace_lines

    def trace_error(self) -> List[str]:
        return List('<erroneous traceback>')

    @property
    def _internal_packages_arg(self) -> Optional[List[str]]:
        return self.internal_packages | None

    @property
    def full_trace_lines(self) -> List[str]:
        return self.trace_lines_with(format_exception(self.cause))

    @property
    def truncated_trace_lines(self) -> List[str]:
        try:
            tb = Lists.wrap(traceback.walk_tb(self.cause.__traceback__)) / _[0]
            error_loc = (
                tb
                .filter(L(non_internal_frame)(_, self._internal_packages_arg))
                .flat_map(frame_traceback_entry)
            )
            return self.trace_lines_with((error_loc) + format_exception_error(self.cause))
        except Exception as e:
            return self.trace_error()

    def trace_lines_with(self, extra: List[str]) -> List[str]:
        cs = callsite_info(self.frame, self._internal_packages_arg)
        return cs.cons(self.desc) + extra

    def __str__(self) -> str:
        return self.lines.join_lines

    @property
    def callsite(self) -> Any:
        return frame_callsite(self.frame)

    @property
    def callsite_source(self) -> List[str]:
        return callsite_source(self.frame)


class IOException(IOExceptionBase):

    @property
    def desc(self) -> str:
        return 'IO exception'

    @property
    def internal_packages(self) -> Maybe[List[str]]:
        return Nothing


class IOMeta(ImplicitsMeta):

    @property
    def zero(self):
        return IO.now(None)


def safe_fmt(f: Callable[..., Any], a: tuple, kw: dict) -> Eval[str]:
    def s() -> str:
        try:
            return format_funcall(f, Lists.wrap(a), kw)
        except Exception as e:
            return str(f)
            log.error(str(e))
    return Eval.later(s)


class IO(Generic[A], Implicits, ToStr, implicits=True, metaclass=IOMeta):
    debug = options.io_debug.exists
    stack_only_location = True

    @staticmethod
    def delay(f: Callable[..., A], *a: Any, **kw: Any) -> 'IO[A]':
        def g() -> IO[A]:
            return Pure(f(*a, **kw))
        return Suspend(g, safe_fmt(f, a, kw))

    @staticmethod
    def suspend(f: Callable[..., 'IO[A]'], *a: Any, **kw: Any) -> 'IO[A]':
        return Suspend(lambda: f(*a, **kw), safe_fmt(f, a, kw))

    @staticmethod
    def call(f: Callable[..., A], *a, **kw):
        return IO.delay(f, *a, **kw)

    @staticmethod
    def now(a: A) -> 'IO[A]':
        return Pure(a)

    @staticmethod
    def pure(a: A) -> 'IO[A]':
        return Pure(a)

    @staticmethod
    def just(a: A) -> 'IO[Maybe[A]]':
        return IO.now(Just(a))

    @staticmethod
    def failed(err: str) -> 'IO[A]':
        def fail():
            raise Exception(err)
        return IO.suspend(fail)

    @staticmethod
    def from_either(a: Either[Any, A]) -> 'IO[A]':
        return a.cata(IO.failed, IO.pure)

    @staticmethod
    def e(a: Either[Any, A]) -> 'IO[A]':
        return IO.from_either(a)

    @staticmethod
    def fork_io(f: Callable[..., 'IO[None]'], *a: Any, daemon: bool=True, **kw: Any) -> 'IO[Thread]':
        def run() -> None:
            try:
                f(*a, **kw).attempt.lmap(lambda a: log.error(f'forked IO failed: {a}'))
            except Exception as e:
                log.caught_exception_error(f'running forked IO', e)
        thread = Thread(target=run, daemon=daemon)
        return IO.delay(thread.start).replace(thread)

    @staticmethod
    def fork(f: Callable[..., None], *a: Any, **kw: Any) -> 'IO[Thread]':
        return IO.fork_io(IO.delay, f, *a, **kw)

    @staticmethod
    def from_maybe(a: Maybe[A], error: str) -> 'IO[A]':
        return a / IO.now | IO.failed(error)

    @staticmethod
    def sleep(duration: float) -> 'IO[None]':
        return IO.delay(time.sleep, duration)

    @staticmethod
    def file(path: Path) -> 'IO[List[str]]':
        return IO.delay(path.read_text).map(Lists.lines)

    def __init__(self) -> None:
        self.stack = inspect.stack() if IO.debug else []

    @abc.abstractmethod
    def _flat_map(self, f: Callable[[A], 'IO[B]'], ts: Eval[str], fs: Eval[str]) -> 'IO[B]':
        ...

    @abc.abstractmethod
    def _step(self) -> 'IO[A]':
        ...

    @abc.abstractmethod
    def lambda_str(self) -> Eval[str]:
        ...

    def _arg_desc(self) -> List[str]:
        return List(self.lambda_str()._value())

    def step(self) -> Union[A, 'IO[A]']:
        try:
            return self._step_timed() if IO.debug else self._step()
        except IOException as e:
            raise e
        except Exception as e:
            raise IOException(self.lambda_str()._value(), self.stack, e)

    def run(self) -> A:
        @tailrec
        def run(t: Union[A, 'IO[A]']) -> Union[Tuple[bool, A], Tuple[bool, Tuple[Union[A, 'IO[A]']]]]:
            if isinstance(t, Pure):
                return True, (t.value,)
            elif isinstance(t, IO):
                return True, (t.step(),)
            else:
                return False, t
        return run(self)

    def flat_map(self, f: Callable[[A], 'IO[B]'], ts: Maybe[Eval[str]]=Nothing, fs: Maybe[Eval[str]]=Nothing
                 ) -> 'IO[B]':
        ts1 = ts | self.lambda_str
        fs1 = fs | Eval.later(lambda: f'flat_map({lambda_str(f)})')
        return self._flat_map(f, ts1, fs1)

    def _step_timed(self) -> Union[A, 'IO[A]']:
        start = time.time()
        v = self._step()
        dur = time.time() - start
        if dur > 0.1:
            log.debug2(lambda: 'IO {} took {:.4f}s'.format(self.string(), dur))
        return v

    def _attempt(self) -> Either[IOException, A]:
        try:
            return Right(self.run())
        except IOException as e:
            return Left(e)

    def attempt_run(self) -> Either[IOException, A]:
        return self._attempt()

    @property
    def attempt(self) -> Either[IOException, A]:
        return self._attempt()

    @property
    def fatal(self) -> A:
        return self.attempt.get_or_raise()

    def and_then(self, nxt: 'IO[B]') -> 'IO[B]':
        fs = Eval.later(lambda: 'and_then({nxt.lambda_str()})')
        return self.flat_map(lambda a: nxt, fs=Just(fs))

    __add__ = and_then

    def join_maybe(self, err: str) -> 'IO[B]':
        def f(a: Union[A, Maybe[B]]) -> 'IO[B]':
            return (
                IO.from_maybe(a, err)
                if isinstance(a, Maybe) else
                IO.failed(f'`IO.join_maybe` called on {self}')
            )
        return self.flat_map(f, fs=Just(Eval.later('join_maybe({})'.format, err)))

    @property
    def join_either(self) -> 'IO[B]':
        def f(a: Union[A, Either[C, B]]) -> 'IO[B]':
            return (
                IO.from_either(a)
                if isinstance(a, Either) else
                IO.failed(f'`IO.join_either` called on {self}')
            )
        return self.flat_map(f, fs=Just(Eval.now('join_either')))

    def with_stack(self, s: List[inspect.FrameInfo]) -> 'IO[A]':
        self.stack = s
        return self

    def recover(self, f: Callable[[IOException], B]) -> 'IO[B]':
        return IO.delay(lambda: self.attempt).map(__.value_or(f))

    def recover_with(self, f: Callable[[IOException], 'IO[B]']) -> 'IO[B]':
        return IO.delay(lambda: self.attempt).flat_map(__.map(IO.pure).value_or(f))

    def ensure(self, f: Callable[[Either[IOException, A]], 'IO[None]']) -> 'IO[A]':
        return io_ensure(self, f)

    @property
    def coro(self) -> Awaitable[Either[IOException, A]]:
        async def coro() -> Either[IOException, A]:
            return self.attempt
        return coro()


class Suspend(Generic[A], IO[A]):

    def __init__(self, thunk: Callable[[], IO[A]], string: Eval[str]) -> None:
        super().__init__()
        self.thunk = thunk
        self.string = string

    def lambda_str(self) -> Eval[str]:
        return self.string

    def _step(self) -> IO[A]:
        return self.thunk().with_stack(self.stack)

    def _flat_map(self, f: Callable[[A], IO[B]], ts: Eval[str], fs: Eval[str]) -> IO[B]:
        return BindSuspend(self.thunk, f, ts, fs)


class BindSuspend(Generic[A], IO[A]):

    def __init__(self, thunk: Callable[[], IO[A]], f: Callable, ts: Eval[str], fs: Eval[str]) -> None:
        super().__init__()
        self.thunk = thunk
        self.f = f
        self.ts = ts
        self.fs = fs

    def lambda_str(self) -> Eval[str]:
        return (self.ts & self.fs).map2('{}.{}'.format)

    def _step(self) -> IO[A]:
        return (
            self.thunk()
            .flat_map(self.f, fs=Just(self.fs))
            .with_stack(self.stack)
        )

    def _flat_map(self, f: Callable[[A], IO[B]], ts: Eval[str], fs: Eval[str]) -> IO[B]:
        bs = L(BindSuspend)(self.thunk, lambda a: self.f(a).flat_map(f, Just(ts), Just(fs)), ts, fs)
        return Suspend(bs, (ts & fs).map2('{}.{}'.format))


class Pure(Generic[A], IO[A]):

    def __init__(self, value: A) -> None:
        super().__init__()
        self.value = value

    def lambda_str(self) -> Eval[str]:
        return Eval.later(lambda: f'Pure({self.value})')

    def _arg_desc(self) -> List[str]:
        return List(str(self.value))

    def _step(self) -> IO[A]:
        return self

    def _flat_map(self, f: Callable[[A], IO[B]], ts: Eval[str], fs: Eval[str]) -> IO[B]:
        return Suspend(L(f)(self.value), (ts & fs).map2('{}.{}'.format))


def io(fun: Callable[..., A]) -> Callable[..., IO[A]]:
    def dec(*a: Any, **kw: Any) -> IO[A]:
        return IO.delay(fun, *a, **kw)
    return dec


@do(IO[A])
def io_ensure(io: IO[A], f: Callable[[Either[IOException, A]], 'IO[None]']) -> Do:
    result = yield IO.delay(lambda: io.attempt)
    yield f(result)
    yield IO.from_either(result)


__all__ = ('IO', 'io')



from typing import TypeVar, Callable

from lenses import bind
from lenses.ui import UnboundLens, BoundLens

from amino import Maybe, List, __, Boolean, _, L
from amino.lenses.lens import lens

A = TypeVar('A')

_add = lambda l: __.map(lens.add_lens(l).add_lens)


def path_lens_pred(a: A, sub: Callable[[A], List[A]], lsub, f: Callable[[A], bool]) -> Maybe[BoundLens]:
    g = lambda a: Boolean(f(lsub(a))).maybe(lsub(lens))
    return path_lens(a, sub, g)


def path_lens(a: A, sub: Callable[[A], List[A]], f: Callable[[A], Maybe[UnboundLens]]) -> Maybe[BoundLens]:
    return _path_lens(a, sub, f).map(lambda b: bind(a).tuple_(*b))


def path_lens_unbound_pre(a: A, sub: Callable[[A], List[A]], f: Callable[[A], Maybe[UnboundLens]], pre: Callable
                          ) -> Maybe[UnboundLens]:
    return (
        _path_lens(pre(a), sub, f) /
        (_ / pre(lens).add_lens) /
        __.cons(lens)
    ).map(lambda b: lens.tuple_(*b))


def path_lens_unbound(a: A, sub: Callable[[A], List[A]], f: Callable[[A], Maybe[UnboundLens]]) -> Maybe[UnboundLens]:
    return _path_lens(a, sub, f).map(lambda b: lens.tuple_(*b))


def _path_lens(a: A, sub: Callable[[A], List[A]], f: Callable[[A], Maybe[UnboundLens]]) -> Maybe[UnboundLens]:
    def go_sub():
        l, s = sub(lens), sub(a)
        g = lambda b: _path_lens(b, sub, f)
        return _path_lens_list(s, g) / _add(l).cons(lens)
    return (f(a) / L(List)(lens, _)).or_else(go_sub)


def _path_lens_list(fa: List[A], f: Callable[[A], Maybe[UnboundLens]]) -> Maybe[UnboundLens]:
    check = lambda a: f(a[1]) / (lambda b: (a[0], b))
    cat = lambda i, l: _add(lens[i])(l)
    return fa.with_index.find_map(check).map2(cat)

__all__ = ('path_lens',)



import itertools
from typing import TypeVar, Callable, Tuple, Optional as TOptional, Type
from functools import reduce

from amino import maybe, List, Maybe, Lists
from amino.func import curried, call_by_name, CallByName
from amino.lazy import lazy
from amino.tc.monad import Monad
from amino.tc.base import ImplicitInstances, tc_prop, F
from amino.tc.traverse import Traverse, TraverseF, TraverseG
from amino.tc.applicative import Applicative
from amino.tc.foldable import Foldable, FoldableABC
from amino.tc.zip import Zip
from amino.tc.monoid import Monoid
from amino.tc.optional import Optional

A = TypeVar('A', covariant=True)
B = TypeVar('B')


class ListInstances(ImplicitInstances):

    @lazy
    def _instances(self):
        from amino import Map
        return Map(
            {
                Monad: ListMonad(),
                Traverse: ListTraverse(),
                Foldable: ListFoldable(),
                Zip: ListZip(),
                Monoid: ListMonoid(),
            }
        )


class ListMonad(Monad):

    def pure(self, b: B) -> List[B]:
        return List(b)

    def map(self, fa: List[A], f: Callable[[A], B]) -> List[B]:
        return List.wrap(map(f, fa))

    def flat_map(self, fa: List[A], f: Callable[[A], List[B]]) -> List[B]:
        acc = []
        for a in fa:
            fb = f(a)
            acc.append(fb if isinstance(fb, List) else fb.to_list)
        return Lists.wrap(itertools.chain.from_iterable(acc))


class ListTraverse(Traverse):

    def traverse(self, fa: List[A], f: Callable, tpe: type):
        monad = Applicative.fatal(tpe)
        def folder(z, a):
            return monad.map2(z.product(f(a)), lambda l, b: l.cat(b))
        return fa.fold_left(monad.pure(List()))(folder)


FoldableABC.register(List)
TraverseF.register(List)
TraverseG.register(List)


def _find(fa: List[A], f: Callable[[A], bool]) -> TOptional[A]:
    return next(filter(f, fa), None)


class ListFoldable(Foldable):

    @tc_prop
    def with_index(self, fa: List[A]) -> List[Tuple[int, A]]:
        return List.wrap(enumerate(fa))

    def filter(self, fa: List[A], f: Callable[[A], bool]):
        return List.wrap(filter(f, fa))

    @curried
    def fold_left(self, fa: List[A], z: B, f: Callable[[B, A], B]) -> B:
        return reduce(f, fa, z)

    def find(self, fa: List[A], f: Callable[[A], bool]):
        return Maybe.optional(_find(fa, f))

    def find_map_optional(self, fa: List[A], tpe: Type[F], f: Callable[[A], F[B]], msg: CallByName=None) -> F[B]:
        for el in fa:
            found = f(el)
            if found.present:
                return found
        return Optional.fatal(tpe).absent(call_by_name(msg))

    def index_where(self, fa: List[A], f: Callable[[A], bool]):
        gen = (maybe.Just(i) for i, a in enumerate(fa) if f(a))
        return next(gen, maybe.Nothing)


class ListZip(Zip):

    def zip(self, fa: List[A], fb: List[B], *fs) -> List:
        return List.wrap(zip(fa, fb, *fs))


class ListMonoid(Monoid):

    @property
    def empty(self):
        return List()

    def combine(self, fa: List, fb: List):
        return fa + fb

__all__ = ('ListInstances',)



import json
from typing import Any

from amino import Either, Lists, Map, Try
from amino.json.data import Json, JsonArray, JsonScalar, JsonObject, JsonNull, JsonError


def to_json(a: Any) -> Json:
    return (
        JsonArray(Lists.wrap(a) / to_json)
        if isinstance(a, (list, tuple)) else
        JsonObject(Map(a).valmap(to_json))
        if isinstance(a, dict) else
        JsonNull(None)
        if a is None else
        JsonScalar(a)
    )


def parse_json(payload: str) -> Either[str, Json]:
    return (Try(json.loads, payload) / to_json).lmap(lambda e: JsonError(payload, e))



from typing import Union, TypeVar, Mapping, Any, List as TList, Callable, Tuple, Type
from numbers import Number
from uuid import UUID
from types import FunctionType

from amino import Either, List, L, _, Right, Lists, Maybe, Path, Map, Boolean, do, Do, Try, Dat
from amino.json.encoder import Encoder, encode_json, json_object_with_type
from amino.json.data import JsonError, Json, JsonArray, JsonScalar, JsonObject, JsonNull

A = TypeVar('A')
B = TypeVar('B')
Sub = TypeVar('Sub', bound=Dat)


class ScalarEncoder(Encoder[Union[Number, str, None]], pred=L(issubclass)(_, (Number, str, type(None)))):

    def encode(self, a: Union[Number, str, None]) -> Either[JsonError, Json]:
        return Right(JsonScalar(a))


class MapEncoder(Encoder[List], pred=L(issubclass)(_, Mapping)):

    def encode(self, a: Map[str, Any]) -> Either[JsonError, Json]:
        return Map(a).traverse(encode_json, Either) / JsonObject


class ListEncoder(Encoder[List], pred=L(issubclass)(_, TList)):

    def encode(self, a: TList) -> Either[JsonError, Json]:
        return Lists.wrap(a).traverse(encode_json, Either) / JsonArray


class MaybeEncoder(Encoder[Maybe], tpe=Maybe):

    def encode(self, a: Maybe[A]) -> Either[JsonError, Json]:
        return a.map(encode_json) | (lambda: Right(JsonNull.cons()))


class EitherEncoder(Encoder[Either], tpe=Either):

    @do(Either[JsonError, Json])
    def encode(self, a: Either[B, A]) -> Do:
        json = yield encode_json(a.value)
        yield Right(json_object_with_type(Map(value=json), type(a)))


class UUIDEncoder(Encoder[UUID], tpe=UUID):

    def encode(self, a: UUID) -> Either[JsonError, Json]:
        return Right(JsonScalar(str(a)))


class PathEncoder(Encoder[Path], tpe=Path):

    def encode(self, a: Path) -> Either[JsonError, Json]:
        return Right(JsonScalar(str(a)))


class BooleanEncoder(Encoder[Boolean], tpe=Boolean):

    def encode(self, a: Boolean) -> Either[JsonError, Json]:
        return Right(JsonScalar(a.value))


@do(Either[JsonError, Json])
def encode_instance(a: A, tpe: type, module: str, names: list) -> Do:
    mod_json = yield encode_json(module)
    names_json = yield encode_json(names)
    return json_object_with_type(Map(module=mod_json, names=names_json), tpe)


@do(Either[JsonError, Json])
def encode_instance_simple(data: A, tpe: type) -> Do:
    mod = yield Try(lambda: data.__module__)
    names = yield Try(lambda: data.__qualname__.split('.'))
    yield encode_instance(data, tpe, mod, names)


class FunctionEncoder(Encoder[Callable], tpe=FunctionType):

    @do(Either[JsonError, Json])
    def encode(self, data: Callable) -> Do:
        yield encode_instance_simple(data, Callable)


class TupleEncoder(Encoder[Tuple], tpe=tuple):

    @do(Either[JsonError, Json])
    def encode(self, data: Tuple) -> Do:
        array = yield encode_json(Lists.wrap(data))
        return json_object_with_type(Map(data=array), tuple)


class TypeEncoder(Encoder[Type], tpe=type):

    def encode(self, data: Type) -> Either[JsonError, Json]:
        return encode_instance_simple(data, Type)


class DatEncoder(Encoder, tpe=Dat):

    @do(Either[JsonError, Map])
    def encode(self, a: Sub) -> Do:
        jsons = yield a._dat__values.traverse(encode_json, Either)
        yield Right(json_object_with_type(Map(a._dat__names.zip(jsons)), type(a)))


__all__ = ('ListEncoder', 'ScalarEncoder', 'MaybeEncoder', 'UUIDEncoder', 'PathEncoder', 'MapEncoder', 'EitherEncoder',
           'BooleanEncoder', 'encode_instance', 'FunctionEncoder', 'TupleEncoder', 'TypeEncoder', 'DatEncoder',)



from typing import Callable, TypeVar, Any, Generic

from amino.test.spec_spec import Spec
from amino.typeclass.typeclass import instance
from amino.typeclass.functor import Functor, fmap

A = TypeVar('A')
B = TypeVar('B')
F = TypeVar('F')

_A = TypeVar('_A')

class TMaybe(Generic[_A]):
    pass


class TJust(TMaybe[_A]):

    def __init__(self, value: _A) -> None:
        self.value = value


class _Nothing(TMaybe[_A]):
    pass


TNothing: TMaybe[Any] = _Nothing()


def fmap_maybe(f: Callable[[A], B], fa: TMaybe[A]) -> TMaybe[B]:
    return TJust(f(fa.value)) if isinstance(fa, TJust) else TNothing


maybe_instance = instance(Functor, TMaybe, fmap=fmap_maybe)


class _TypeclassSpec(Spec):

    def fmap(self) -> None:
        print(fmap(lambda a: a + 1)(TJust(1)))


__all__ = ('_TypeclassSpec',)



import re
import abc
from typing import Callable, Any, List, TypeVar

from amino.tc.base import TypeClass

A = TypeVar('A')
B = TypeVar('B')
F = TypeVar('F')


def apply_n(self: TypeClass, num: int, fa: F, f: Callable[..., A], g: Callable[..., B], *a: Any) -> B:
    def wrapper(args: tuple):
        if len(args) != num:
            name = self.__class__.__name__
            msg = f'passed {len(args)} args to {name}.{g.__name__}{num}'
            raise TypeError(msg)
        return f(*args)
    return g(fa, wrapper, *a)


apply_n_rex = re.compile('^(.*)(\d+)$')


class ApplyN:

    @abc.abstractclassmethod
    def apply_n_funcs(self) -> List[str]:
        ...

    def __getattr__(self, name: str) -> Callable:
        def error() -> None:
            raise AttributeError(f'''`{self.__class__.__name__}` object has no attribute `{name}`''')
        match = apply_n_rex.match(name)
        if match is None:
            error()
        func = match.group(1)
        num = int(match.group(2))
        n_name = f'{func}_n'
        if hasattr(self, n_name):
            return lambda fa, f, *a: getattr(self, n_name)(num, fa, f, *a)
        if func not in self.apply_n_funcs():
            error()
        return lambda fa, f, *a: apply_n(self, num, fa, f, getattr(self, func), *a)


__all__ = ('apply_n', 'ApplyN')



from typing import TypeVar, Callable, Union, Type, Tuple

from amino.tc.base import tc_prop, ImplicitInstances, F
from amino.tc.monad import Monad
from amino.tc.optional import Optional
from amino import either, Just, Empty, Maybe, curried, List, Nothing
from amino.lazy import lazy
from amino.maybe import call_by_name
from amino.tc.applicative import Applicative
from amino.tc.traverse import Traverse
from amino.tc.foldable import Foldable
from amino.tc.zip import Zip
from amino.instances.list import ListTraverse
from amino.func import CallByName

A = TypeVar('A')
B = TypeVar('B')
C = TypeVar('C')


class MaybeInstances(ImplicitInstances):

    @lazy
    def _instances(self):
        from amino import Map
        return Map(
            {
                Monad: MaybeMonad(),
                Optional: MaybeOptional(),
                Traverse: MaybeTraverse(),
                Foldable: MaybeFoldable(),
                Zip: MaybeZip(),
            }
        )


class MaybeMonad(Monad):

    def pure(self, a: A):
        return Just(a)

    def flat_map(self, fa: Maybe[A], f: Callable[[A], Maybe[B]]) -> Maybe[B]:
        return fa.cata(lambda v: f(v), Nothing)


class MaybeOptional(Optional):

    @tc_prop
    def to_maybe(self, fa: Maybe[A]):
        return fa

    def to_either(self, fa: Maybe[A], left: Union[B, Callable[[], B]]
                  ) -> either.Either[A, B]:
        from amino.either import Left, Right
        return fa.cata(Right, lambda: Left(call_by_name(left)))

    @tc_prop
    def present(self, fa: Maybe):
        return fa.is_just

    def absent(self, msg: str) -> Maybe[A]:
        return Nothing


class MaybeTraverse(Traverse):

    def traverse(self, fa: Maybe[A], f: Callable, tpe: type):
        monad = Applicative.fatal(tpe)
        r = lambda a: monad.map(f(a), Just)
        return fa.cata(r, monad.pure(Empty()))


class MaybeFoldable(Foldable):

    @tc_prop
    def with_index(self, fa: Maybe[A]) -> Maybe[Tuple[int, A]]:
        return Just(0) & fa

    def filter(self, fa: Maybe[A], f: Callable[[A], bool]):
        return fa // (lambda a: Just(a) if f(a) else Empty())

    @curried
    def fold_left(self, fa: Maybe[A], z: B, f: Callable[[B, A], B]) -> B:
        return fa / (lambda a: f(z, a)) | z

    def find(self, fa: Maybe[A], f: Callable[[A], bool]):
        return self.filter(fa, f)

    def find_map_optional(self, fa: Maybe[A], tpe: Type[F], f: Callable[[B], F[C]], msg: CallByName=None) -> F[C]:
        return fa / f | (lambda: fa.absent(call_by_name(msg)))

    def index_where(self, fa: Maybe[A], f: Callable[[A], bool]):
        return fa / f // (lambda a: Just(0) if a else Empty())


class MaybeZip(Zip):

    def zip(self, fa: Maybe[A], fb: Maybe[B], *fs: Maybe) -> Maybe:
        return ListTraverse().sequence(List(fa, fb, *fs), Maybe)

__all__ = ('MaybeInstances',)



from .path import setup, create_temp_file, temp_file, temp_path, temp_dir, fixture_path, load_fixture

__all__ = ('setup', 'create_temp_file', 'temp_file', 'temp_path', 'temp_dir', 'fixture_path', 'load_fixture')



import abc
from typing import TypeVar, Generic, Callable, List

from amino.tc.base import TypeClass
from amino.func import I
from amino import _
from amino.tc.apply_n import ApplyN

A = TypeVar('A')
B = TypeVar('B')


class TraverseF(Generic[A], abc.ABC):
    pass


class TraverseG(Generic[A], abc.ABC):
    pass

F = TraverseF
G = TraverseG

F0 = TypeVar('F0', bound=TraverseF)
G0 = TypeVar('G0', bound=TraverseG)


class Traverse(TypeClass, ApplyN):
    # FIXME lens functions return index lenses, which is not a property of Traverse

    def apply_n_funcs(self) -> list:
        return ['traverse', 'flat_traverse']

    @abc.abstractmethod
    def traverse(self, fa: F[G[A]], f: Callable[[A], B], tpe: type) -> G[F[B]]:
        ...

    def flat_traverse(self, fa: F[G[A]], f: Callable[[A], F[B]], tpe: type) -> G[F[B]]:
        return self.traverse(fa, f, tpe).map(_.join)  # type: ignore

    def sequence(self, fa: F[G[A]], tpe: type) -> G[F[A]]:
        return self.traverse(fa, I, tpe)

    def flat_sequence(self, fa: F[G[A]], tpe: type) -> G[F[B]]:
        return self.sequence(fa, tpe).map(_.join)  # type: ignore

__all__ = ('Traverse',)



from typing import Iterable, TypeVar, Any

from lenses.ui import UnboundLens
from lenses.optics import TrivialIso
from lenses.hooks.hook_funcs import from_iter

from amino import List, Lists


class UnboundLensA(UnboundLens):

    def __getattr__(self, name):
        # type: (str) -> Any
        if name.startswith('__') and name.endswith('__') or name.startswith('call_mut_') or name.startswith('call_'):
            return super().__getattr__(name)
        else:
            return self.GetAttr(name)


lens = UnboundLensA(TrivialIso())

A = TypeVar('A')


def list_from_iter(self, iterable: Iterable[A]) -> List[A]:
    return Lists.wrap(iterable)


from_iter.register(list)(list_from_iter)

__all__ = ('lens',)



import abc
import importlib
from typing import TypeVar, Generic, Union, Any, cast, Iterator, Type, Callable
from types import ModuleType
from pathlib import Path

import amino
from amino import boolean
from amino.func import I
from amino.tc.base import F
from amino.util.mod import unsafe_import_name
from amino.tc.monoid import Monoid
from amino.tc.monad import Monad
from amino.util.string import ToStr
from amino.do import do, Do
from amino.util.exception import format_exception

A = TypeVar('A')
B = TypeVar('B')
C = TypeVar('C')
D = TypeVar('D')
E = TypeVar('E', bound=Exception)


class ImportFailure(ToStr):

    @abc.abstractproperty
    def expand(self) -> 'amino.list.List[str]':
        ...


class ImportException(ImportFailure):

    def __init__(self, desc: str, exc: Exception) -> None:
        self.desc = desc
        self.exc = exc

    def _arg_desc(self) -> 'amino.list.List':
        return amino.List(self.desc, str(self.exc))

    @property
    def expand(self) -> 'amino.list.List':
        return format_exception(self.exc).cons(self.desc)


class InvalidLocator(ImportFailure):

    def __init__(self, msg: str) -> None:
        self.msg = msg

    def _arg_desc(self) -> 'amino.list.List':
        return amino.List(self.msg)

    @property
    def expand(self) -> 'amino.list.List':
        return amino.List(self.msg)


class Either(Generic[A, B], F[B], implicits=True):

    @staticmethod
    def import_name(mod: str, name: str) -> 'Either[ImportFailure, B]':
        try:
            value = (
                __builtins__[name]
                if mod == '__builtins__' else
                unsafe_import_name(mod, name)
            )
        except Exception as e:
            return Left(ImportException(f'{mod}.{name}', e))
        else:
            return Left(InvalidLocator(f'{mod} has no attribute {name}')) if value is None else Right(value)

    @staticmethod
    def import_path(path: str) -> 'Either[ImportFailure, B]':
        from amino.list import List
        return (
            List.wrap(path.rsplit('.', 1))
            .lift_all(0, 1)
            .to_either(InvalidLocator(f'invalid module path: {path}'))
            .flat_map2(lambda a, b: Either.import_name(a, b).lmap(lambda c: ImportException(path, c)))
        )

    @staticmethod
    def import_module(modname: str) -> 'Either[ImportFailure, ModuleType]':
        try:
            mod = importlib.import_module(modname)
        except Exception as e:
            return Left(ImportException(modname, e))
        else:
            return Right(mod)

    @staticmethod
    def import_file(path: Path) -> 'Either[ImportFailure, ModuleType]':
        from amino.maybe import Maybe
        def step2(spec: importlib._bootstrap.ModuleSpec) -> ModuleType:
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)
            return module
        try:
            module = Maybe.check(importlib.util.spec_from_file_location('temp', str(path))) / step2
        except Exception as e:
            return Left(ImportException(str(path), e))
        else:
            return module.to_either(InvalidLocator(f'failed to import `{path}`'))

    @staticmethod
    def import_from_file(path: Path, name: str) -> 'Either[ImportFailure, B]':
        @do(Either[ImportFailure, B])
        def run() -> Do:
            module = yield Either.import_file(path)
            attr = getattr(module, name, None)
            yield (
                Left(InvalidLocator(f'{path} has no attribute {name}'))
                if attr is None else
                Right(attr)
            )
        return run()

    @staticmethod
    def exports(modpath: str) -> 'Either[ImportFailure, amino.List[Any]]':
        @do(Either[ImportFailure, amino.List[Any]])
        def run() -> Do:
            from amino.list import Lists
            exports = yield Either.import_name(modpath, '__all__')
            yield Lists.wrap(exports).traverse(lambda a: Either.import_name(modpath, a), Either)
        return run()

    @staticmethod
    def getattr(obj: Any, attr: str) -> 'Either[str, A]':
        return Right(getattr(obj, attr)) if hasattr(obj, attr) else Left(f'`{obj}` has no attribute `{attr}`')

    @property
    def is_right(self) -> 'amino.boolean.Boolean':
        return boolean.Boolean(isinstance(self, Right))

    @property
    def is_left(self) -> 'amino.boolean.Boolean':
        return boolean.Boolean(isinstance(self, Left))

    @property
    def __left_value(self) -> A:
        return cast(A, self.value)

    @property
    def __right_value(self) -> B:
        return cast(B, self.value)

    def leffect(self, f: Callable[[A], Any]) -> 'Either[A, B]':
        if self.is_left:
            f(self.__left_value)
        return self

    def bieffect(self, l: Callable[[A], Any], r: Callable[[B], Any]) -> 'Either[A, B]':
        self.cata(l, r)
        return self

    def cata(self, fl: Callable[[A], C], fr: Callable[[B], C]) -> C:
        return fl(self.__left_value) if self.is_left else fr(self.__right_value)

    def bimap(self, fl: Callable[[A], C], fr: Callable[[B], D]) -> 'Either[C, D]':
        return Left(fl(self.__left_value)) if self.is_left else Right(fr(self.__right_value))

    def recover_with(self, f: Callable[[A], 'Either[C, B]']) -> 'Either[C, B]':
        return self.cata(f, Right)

    def right_or_map(self, f: Callable[[A], C]) -> C:
        return self.cata(f, I)

    def value_or(self, f: Callable[[A], B]) -> B:
        return self.cata(f, I)

    def left_or(self, f: Callable[[B], A]) -> A:
        return self.cata(I, f)

    def left_or_map(self, f: Callable[[B], A]) -> A:
        return self.cata(I, f)

    @property
    def ljoin(self) -> 'Either[A, C]':
        return self.right_or_map(Left)

    def __str__(self) -> str:
        return '{}({})'.format(self.__class__.__name__, self.value)

    def __repr__(self) -> str:
        return '{}({!r})'.format(self.__class__.__name__, self.value)

    @property
    def to_list(self) -> 'amino.list.List[B]':
        return self.to_maybe.to_list

    def lmap(self, f: Callable[[A], C]) -> 'Either[C, B]':
        return cast(Either, Left(f(self.__left_value))) if self.is_left else cast(Either, Right(self.__right_value))

    def get_or_raise(self) -> B:
        def fail(err: A) -> B:
            raise err if isinstance(err, Exception) else Exception(err)
        return self.cata(fail, I)

    @property
    def fatal(self) -> B:
        return self.get_or_raise()

    def __iter__(self) -> Iterator[B]:
        return iter(self.to_list)

    @property
    def swap(self) -> 'Either[B, A]':
        return self.cata(Right, Left)

    @property
    def json_repr(self) -> B:
        return self.to_maybe.json_repr

    def accum_error(self, b: 'Either[A, C]') -> 'Either[A, C]':
        return self.accum_error_f(lambda: b)

    def accum_error_f(self, f: Callable[[], 'Either[A, C]'], *a, **kw) -> 'Either[A, C]':
        def acc(v: A) -> 'Either[A, C]':
            monoid = Monoid.fatal_for(self.__left_value)
            return monoid.combine(self.__left_value, v)
        return f(*a, **kw).lmap(acc) if self.is_left else self

    def accum_error_lift(self, f: Callable[[], 'Either[A, C]'], *a, **kw) -> 'Either[A, C]':
        def acc(v: A) -> 'Either[A, C]':
            monoid = Monoid.fatal_for(self.__left_value)
            monad = Monad.fatal_for(self.__left_value)
            return monoid.combine(self.__left_value, monad.pure(v))
        return f(*a, **kw).lmap(acc) if self.is_left else self

    def filter_with(self, f: Callable[[B], bool], g: Callable[[B], C]) -> 'Either[C, B]':
        return self // (lambda a: Right(a) if f(a) else Left(g(a)))

    def left_contains(self, a: A) -> 'amino.boolean.Boolean':
        return boolean.Boolean(self.is_left and self.__left_value == a)


class Right(Either[A, B]):

    def __init__(self, value: B) -> None:
        self.value = value

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, Right) and self._Either__right_value == other._Either__right_value


class Left(Either[A, B]):

    def __init__(self, value: A) -> None:
        self.value = value

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, Left) and self._Either__left_value == other._Either__left_value


def Try(f: Callable[..., A], *a: Any, **kw: Any) -> Either[Exception, A]:
    try:
        return Right(f(*a, **kw))
    except Exception as e:
        return Left(e)


__all__ = ('Either', 'Left', 'Right', 'ImportFailure', 'ImportException', 'InvalidLocator', 'Try')



from typing import TypeVar, Callable

from amino.tc.monad import Monad
from amino.tc.base import ImplicitInstances
from amino.lazy import lazy
from amino import Map
from amino.compute.compute import Compute, ComputePure, flat_map_compute

A = TypeVar('A')
B = TypeVar('B')
S = TypeVar('S')


class ComputeInstances(ImplicitInstances):

    @lazy
    def _instances(self) -> Map:
        return Map({Monad: ComputeMonad()})


class ComputeMonad(Monad[Compute]):

    def pure(self, a: A) -> Compute[S, A]:
        return ComputePure(a)

    def flat_map(self, fa: Compute[S, A], f: Callable[[A], Compute[S, B]]) -> Compute[S, B]:
        return flat_map_compute(f)(fa)


__all__ = ('ComputeInstances', 'ComputeMonad')



import abc
from typing import Any, Generic, TypeVar
from types import SimpleNamespace

from amino import List, Lists, Nil, Maybe
from amino.util.string import ToStr

A = TypeVar('A')


def is_algebra(bases: List[type]) -> bool:
    return bases.exists(lambda a: hasattr(a, '__algebra_base__'))


def find_algebra(name: str, bases: List[type]) -> Maybe[type]:
    return bases.find(lambda a: hasattr(a, '__algebra_variants__'))


def setup_algebra(name: str, inst: type, bases: List[type]) -> None:
    if is_algebra(bases):
        inst.__algebra_variants__ = List()
    else:
        raise Exception(f'algebra subclass has no algebra superclass: {name}')


def setup_variant(name: str, inst: type, bases: List[type], algebra: type) -> None:
    inst.__algebra_index__ = len(algebra.__algebra_variants__)
    algebra.__algebra_variants__.append(inst)


def setup_algebraic_type(name: str, inst: type, bases: List[type]) -> None:
    return (
        find_algebra(name, bases)
        .cata_f(
            lambda a: setup_variant(name, inst, bases, a),
            lambda: setup_algebra(name, inst, bases)
        )
    )


class AlgebraMeta(abc.ABCMeta):

    def __new__(
            cls,
            name: str,
            bases: list,
            namespace: SimpleNamespace,
            algebra_base: bool=False,
            **kw: Any,
    ) -> None:
        inst = super().__new__(cls, name, bases, namespace, **kw)
        if not hasattr(inst, '__args__') or inst.__args__ is None:
            if algebra_base:
                inst.__algebra_base__ = None
            else:
                setup_algebraic_type(name, inst, Lists.wrap(bases))
        return inst


class Algebra(Generic[A], ToStr, metaclass=AlgebraMeta, algebra_base=True):
    pass


__all__ = ('AlgebraMeta', 'Algebra')



from typing import Generic, TypeVar, Callable, Tuple, cast, Type, Any

from lenses import UnboundLens

from amino.tc.base import ImplicitsMeta, Implicits
from amino.tc.monad import Monad
from amino.tc.zip import Zip
from amino.instances.list import ListTraverse
from amino import List, curried
from amino.util.string import ToStr
from amino.state.base import StateT
from amino.id import Id



A = TypeVar('A')
B = TypeVar('B')
S = TypeVar('S')
R = TypeVar('R')
ST1 = TypeVar('ST1')


monad: Monad = cast(Monad, Monad.fatal(Id))


class IdStateCtor(Generic[S]):

    def inspect(self, f: Callable[[S], A]) -> 'IdState[S, A]':
        def g(s: S) -> Id[Tuple[S, A]]:
            return monad.pure((s, f(s)))
        return IdState.apply(g)

    def inspect_f(self, f: Callable[[S], Id[A]]) -> 'IdState[S, A]':
        def g(s: S) -> Id[Tuple[S, A]]:
            return f(s).map(lambda a: (s, a))
        return IdState.apply(g)

    def pure(self, a: A) -> 'IdState[S, A]':
        return IdState.apply(lambda s: monad.pure((s, a)))

    def delay(self, fa: Callable[..., A], *a: Any, **kw: Any) -> 'IdState[S, A]':
        return IdState.apply(lambda s: monad.pure((s, fa(*a, **kw))))

    def lift(self, fa: Id[A]) -> 'IdState[S, A]':
        def g(s: S) -> Id[Tuple[S, A]]:
            return fa.map(lambda a: (s, a))
        return IdState.apply(g)

    def modify(self, f: Callable[[S], S]) -> 'IdState[S, None]':
        return IdState.apply(lambda s: monad.pure((f(s), None)))

    def modify_f(self, f: Callable[[S], Id[S]]) -> 'IdState[S, None]':
        return IdState.apply(lambda s: f(s).map(lambda a: (a, None)))

    def get(self) -> 'IdState[S, S]':
        return self.inspect(lambda a: a)

    @property
    def unit(self) -> 'IdState[S, None]':
        return IdState.pure(None)



class IdStateMeta(ImplicitsMeta):

    def cons(self, run_f: Id[Callable[[S], Id[Tuple[S, A]]]]) -> 'IdState[S, A]':
        return self(run_f)

    def apply(self, f: Callable[[S], Id[Tuple[S, A]]]) -> 'IdState[S, A]':
        return self.cons(monad.pure(f))

    def apply_f(self, run_f: Id[Callable[[S], Id[Tuple[S, A]]]]) -> 'IdState[S, A]':
        return self.cons(run_f)

    def inspect(self, f: Callable[[S], A]) -> 'IdState[S, A]':
        def g(s: S) -> Id[Tuple[S, A]]:
            return monad.pure((s, f(s)))
        return self.apply(g)

    def inspect_f(self, f: Callable[[S], Id[A]]) -> 'IdState[S, A]':
        def g(s: S) -> Id[Tuple[S, A]]:
            return f(s).map(lambda a: (s, a))
        return self.apply(g)

    def pure(self, a: A) -> 'IdState[S, A]':
        return self.apply(lambda s: monad.pure((s, a)))

    def reset(self, s: S, a: A) -> 'IdState[S, A]':
        return self.apply(lambda _: monad.pure((s, a)))

    def reset_t(self, t: Tuple[S, A]) -> 'IdState[S, A]':
        return self.apply(lambda _: monad.pure(t))

    def delay(self, fa: Callable[..., A], *a: Any, **kw: Any) -> 'IdState[S, A]':
        return self.apply(lambda s: monad.pure((s, fa(*a, **kw))))

    def lift(self, fa: Id[A]) -> 'IdState[S, A]':
        def g(s: S) -> Id[Tuple[S, A]]:
            return fa.map(lambda a: (s, a))
        return self.apply(g)

    def modify(self, f: Callable[[S], S]) -> 'IdState[S, None]':
        return self.apply(lambda s: monad.pure((f(s), None)))

    def modify_f(self, f: Callable[[S], Id[S]]) -> 'IdState[S, None]':
        return self.apply(lambda s: f(s).map(lambda a: (a, None)))

    def set(self, s: S) -> 'IdState[S, None]':
        return self.modify(lambda s0: s)

    def get(self) -> 'IdState[S, S]':
        return self.inspect(lambda a: a)

    @property
    def unit(self) -> 'IdState[S, None]':
        return IdState.pure(None)

    def s(self, tpe: Type[S]) -> IdStateCtor[S]:
        return IdStateCtor()



class IdState(Generic[S, A], StateT, ToStr, Implicits, implicits=True, auto=True, metaclass=IdStateMeta):

    def __init__(self, run_f: Id[Callable[[S], Id[Tuple[S, A]]]]) -> None:
        self.run_f = run_f

    @property
    def cls(self) -> Type['IdState[S, A]']:
        return type(self)

    def run(self, s: S) -> Id[Tuple[S, A]]:
        return self.run_f.flat_map(lambda f: f(s))

    def run_s(self, s: S) -> Id[S]:
        return self.run(s).map(lambda a: a[0])

    def run_a(self, s: S) -> Id[S]:
        return self.run(s).map(lambda a: a[1])

    def _arg_desc(self) -> List[str]:
        return List(str(self.run_f))

    def flat_map_f(self, f: Callable[[A], Id[B]]) -> 'IdState[S, B]':
        def h(s: S, a: A) -> Id[Tuple[S, B]]:
            return f(a).map(lambda b: (s, b))
        def g(fsa: Id[Tuple[S, A]]) -> Id[Tuple[S, B]]:
            return fsa.flat_map2(h)
        run_f1 = self.run_f.map(lambda sfsa: lambda a: g(sfsa(a)))
        return self.cls.apply_f(run_f1)

    def transform(self, f: Callable[[Tuple[S, A]], Tuple[S, B]]) -> 'IdState[S, B]':
        def g(fsa: Id[Tuple[S, A]]) -> Id[Tuple[S, B]]:
            return fsa.map2(f)
        run_f1 = self.run_f.map(lambda sfsa: lambda a: g(sfsa(a)))
        return self.cls.apply_f(run_f1)

    def transform_s(self, f: Callable[[R], S], g: Callable[[R, S], R]) -> 'IdState[R, A]':
        def trans(sfsa: Callable[[S], Id[Tuple[S, A]]], r: R) -> Id[Tuple[R, A]]:
            s = f(r)
            return sfsa(s).map2(lambda s, a: (g(r, s), a))
        return self.cls.apply_f(self.run_f.map(curried(trans)))

    def transform_f(self, tpe: Type[ST1], f: Callable[[Id[Tuple[S, A]]], Any]) -> ST1:
        def trans(s: S) -> Any:
            return f(self.run(s))
        return tpe.apply(trans)  # type: ignore

    def zoom(self, l: UnboundLens) -> 'IdState[R, A]':
        return self.transform_s(l.get(), lambda r, s: l.set(s)(r))

    transform_s_lens = zoom

    def read_zoom(self, l: UnboundLens) -> 'IdState[R, A]':
        return self.transform_s(l.get(), lambda r, s: r)

    transform_s_lens_read = read_zoom

    def flat_map(self, f: Callable[[A], 'IdState[S, B]']) -> 'IdState[S, B]':
        return Monad_IdState.flat_map(self, f)


def run_function(s: IdState[S, A]) -> Id[Callable[[S], Id[Tuple[S, A]]]]:
    try:
        return s.run_f
    except Exception as e:
        if not isinstance(s, IdState):
            raise TypeError(f'flatMapped {s} into IdState')
        else:
            raise


class IdStateMonad(Monad, tpe=IdState):

    def pure(self, a: A) -> IdState[S, A]:  # type: ignore
        return IdState.pure(a)

    def flat_map(  # type: ignore
            self,
            fa: IdState[S, A],
            f: Callable[[A], IdState[S, B]]
    ) -> IdState[S, B]:
        def h(s: S, a: A) -> Id[Tuple[S, B]]:
            return f(a).run(s)
        def g(fsa: Id[Tuple[S, A]]) -> Id[Tuple[S, B]]:
            return fsa.flat_map2(h)
        def i(sfsa: Callable[[S], Id[Tuple[S, A]]]) -> Callable[[S], Id[Tuple[S, B]]]:
            return lambda a: g(sfsa(a))
        run_f1 = run_function(fa).map(i)
        return IdState.apply_f(run_f1)


Monad_IdState = IdStateMonad()


class IdStateZip(Zip, tpe=IdState):

    def zip(
            self,
            fa: IdState[S, A],
            fb: IdState[S, A],
            *fs: IdState[S, A],
    ) -> IdState[S, List[A]]:
        v = ListTraverse().sequence(List(fa, fb, *fs), IdState)  # type: ignore
        return cast(IdState[S, List[A]], v)



__all__ = ('IdState',)



from typing import Callable, TypeVar

from amino.typeclass.typeclass import Class, Signature, Poly, HK, Nullary, HK1, ClassFunction2, Wildcard

A = TypeVar('A')
B = TypeVar('B')

fmap: ClassFunction2 = ClassFunction2(
    'fmap',
    Signature(
        [
            Poly(HK('F', [Nullary('A')]))
        ],
        Poly(HK('F', [Nullary('B')]))
    ),
    Callable[[A], B],
    HK1[A],
    HK1[B],
)

Functor = Class.cons(
    'Functor',
    params=[Poly(HK('F', [Wildcard()]))],
    functions=[
        fmap,
    ]
)

__all__ = ('Functor',)



from typing import Any, Callable

from amino.anon.prod.method import Opers, Anon, AnonMethodCall


class AttrLambda(Opers, Anon):

    def __init__(self, pre: str) -> None:
        self.__pre = pre

    def __pre__(self) -> str:
        return self.__pre

    def __getattr__(self, name: str) -> 'AttrLambda':
        return AttrLambda(f'{self.__pre}.{name}')

    def __getitem__(self, key: Any) -> AnonMethodCall:
        return AnonMethodCall(f'{self.__pre}.__getitem__', (key,), {})

    def __repr__(self) -> str:
        return f'lambda a: {self.__pre}'

    def __call__(self, a: Any) -> Any:
        return self.__eval__()(a)

    def __substitute_object__(self, obj: Any) -> Callable:
        return self.__call__(obj)

    def __lop__(self, op: Callable, s: str, a: Any) -> AnonMethodCall:
        return AnonMethodCall(f'(lambda b: {self.__pre__()} {s} b)', (a,), {})

    def __rop__(self, op: Callable[[Any, Any], Any], s: str, a: Any) -> AnonMethodCall:
        return AnonMethodCall(f'(lambda b: b {s} {self.__pre__()})', (a,), {})

AttrLambdaInst = AttrLambda('a')

__all__ = ('AttrLambdaInst',)



import time
import inspect
from contextlib import contextmanager
from typing import Iterator, cast
from types import FrameType

from amino.logging import module_log

log = module_log()


@contextmanager
def timed() -> Iterator[None]:
    mod = inspect.currentframe()
    caller = cast(FrameType, mod).f_back.f_back
    name = caller.f_globals['__name__']
    start = time.time()
    yield
    total = time.time() - start
    log.info(f'>> {name}: {total}')


__all__ = ('timed',)



from types import FunctionType, MethodType
from typing import Callable, Union


def lambda_str(f: Union[Callable, str]) -> str:
    if isinstance(f, MethodType):
        return '{}.{}'.format(f.__self__.__class__.__name__, f.__name__)
    elif isinstance(f, FunctionType):
        return f.__name__
    elif isinstance(f, str):
        return f
    else:
        return str(f)


def format_funcall(fun: Union[Callable, str], args: tuple, kwargs: dict) -> str:
    from amino.map import Map
    kw = Map(kwargs).map2('{}={!r}'.format)
    a = list(map(repr, args)) + list(kw)
    args_fmt = ', '.join(a)
    return '{}({})'.format(lambda_str(fun), args_fmt)

__all__ = ('lambda_str', 'format_funcall')



from typing import Tuple, TypeVar, Callable

from amino import Maybe, Just, Nothing

A = TypeVar('A')


def lift_tuple(index: int) -> Callable[[Tuple[A, ...]], Maybe[A]]:
    def lift_tuple(data: Tuple[A, ...]) -> Maybe[A]:
        return Just(data[index]) if len(data) > index else Nothing
    return lift_tuple


__all__ = ('lift_tuple',)



from amino.state.maybe import MaybeState
from amino.state.either import EitherState
from amino.state.io import IOState
from amino.state.eval import EvalState
from amino.state.id import IdState
from amino.state.base import StateT

State = IdState

__all__ = ('MaybeState', 'EitherState', 'IOState', 'EvalState', 'IdState', 'StateT', 'State',)



import abc
from typing import TypeVar, Callable, Any, List

from amino.tc.base import TypeClass, tc_prop
from amino.func import ReplaceVal
from amino.tc.apply_n import ApplyN

F = TypeVar('F')
A = TypeVar('A')
B = TypeVar('B')


class Functor(TypeClass, ApplyN):

    def apply_n_funcs(self) -> List[str]:
        return ['map']

    @abc.abstractmethod
    def map(self, fa: F, f: Callable[[A], B]) -> F:
        ...

    def __truediv__(self, fa, f):
        return self.map(fa, f)

    def replace(self, fa: F, b: B) -> F:
        return self.map(fa, ReplaceVal(b))

    @tc_prop
    def void(self, fa: F) -> F:
        return self.replace(fa, None)

    def foreach(self, fa: F, f: Callable[[A], Any]) -> F:
        def effect(a: A) -> A:
            f(a)
            return a
        return self.map(fa, effect)

    __mod__ = foreach

__all__ = ('Functor',)



import random
import string
from functools import reduce
from typing import TypeVar, Callable, Iterable, Any, Union, Tuple, Type, List as TList, Set

from toolz.itertoolz import cons, groupby

from amino import maybe, boolean, Path

from amino.tc.base import ImplicitsMeta, Implicits
from amino.func import curried, I, call_by_name, tailrec, TailrecResult
from amino.util.string import safe_string

A = TypeVar('A')
B = TypeVar('B')


class ListMeta(ImplicitsMeta):

    def __instancecheck__(self, instance: Any) -> bool:
        if type(instance) is list:
            return False
        else:
            return super().__instancecheck__(instance)


def _rand_str(chars: str, num: int=10) -> str:
    return ''.join(random.choice(chars) for i in range(num))


class List(TList[A], Implicits, implicits=True, metaclass=ListMeta):

    def __init__(self, *elements: A) -> None:
        list.__init__(self, elements)

    def __getitem__(self, arg):  # type: ignore
        s = super().__getitem__(arg)
        return List.wrap(s) if isinstance(arg, slice) else s

    @staticmethod
    def wrap(l: Iterable[B]) -> 'List[B]':
        return List(*list(l))

    @staticmethod
    def range(*a: int) -> 'List[int]':
        return List.wrap(range(*a))

    @staticmethod
    def random_string(num: int=10) -> str:
        chars = string.ascii_letters + string.digits
        return _rand_str(chars, num)

    @staticmethod
    def random_alpha(num: int=10) -> str:
        chars = string.ascii_letters
        return _rand_str(chars, num)

    @staticmethod
    def gen(num: int, f: Callable[[], 'List[A]']) -> 'List[A]':
        return List.range(num) // (lambda a: f())

    @staticmethod
    def lines(data: str) -> 'List[str]':
        return List.wrap(data.splitlines())

    def lift(self, index: int) -> 'maybe.Maybe[A]':
        return (
            (maybe.Just(self[index]) if len(self) > index else maybe.Empty())
            if index >= 0 else
            (maybe.Just(self[index]) if len(self) >= -index else maybe.Empty())
        )

    def lift_all(self, index: int, *indexes: int) -> 'maybe.Maybe[List[A]]':
        def folder(z: maybe.Maybe[List[A]], n: List[maybe.Maybe[A]]) -> maybe.Maybe[List[A]]:
            return n.ap(z / (lambda a: a.cat))
        els = List.wrap(indexes) / self.lift
        init = self.lift(index) / List
        return els.fold_left(init)(folder)

    def foreach(self, f: Callable[[A], B]) -> None:
        for el in self:
            f(el)

    def forall(self, f: Callable[[A], bool]) -> 'boolean.Boolean':
        return boolean.Boolean(all(f(el) for el in self))

    def contains(self, value: A) -> 'boolean.Boolean':
        return boolean.Boolean(value in self)

    def exists(self, f: Callable[[A], bool]) -> bool:
        return self.find(f).is_just

    @property
    def is_empty(self) -> 'boolean.Boolean':
        return boolean.Boolean(self.length == 0)

    empty = is_empty

    @property
    def nonempty(self) -> 'boolean.Boolean':
        return not self.empty

    @property
    def length(self) -> int:
        return len(self)

    @property
    def head(self) -> 'maybe.Maybe[A]':
        return self.lift(0)

    @property
    def last(self) -> 'maybe.Maybe[A]':
        return self.lift(-1)

    @property
    def init(self) -> 'maybe.Maybe[List[A]]':
        return maybe.Empty() if self.empty else maybe.Just(self[:-1])

    @property
    def tail(self) -> 'maybe.Maybe[List[A]]':
        return maybe.Empty() if self.empty else maybe.Just(self[1:])

    @property
    def uncons(self) -> 'maybe.Maybe[Tuple[A, List[A]]]':
        return self.head.product(self.tail)

    detach_head = uncons

    @property
    def detach_last(self) -> 'maybe.Maybe[Tuple[A, List[A]]]':
        return self.last.product(self.init)

    @property
    def distinct(self) -> 'List[A]':
        return self.distinct_by(I)

    def distinct_by(self, f: Callable[[A], B]) -> 'List[A]':
        seen: Set[B] = set()
        def pred(a: A) -> bool:
            v = f(a)
            if v in seen:
                return True
            else:
                seen.add(v)
                return False
        return List.wrap(x for x in self if not pred(x))

    def add(self, other: TList[A]) -> 'List[A]':
        return List.wrap(list.__add__(self, other))

    __add__ = add

    def without(self, el: A) -> 'List[A]':
        return self.filter(lambda a: a != el)

    __sub__ = without

    def split(self, f: Callable[[A], bool]) -> Tuple['List[A]', 'List[A]']:
        def splitter(z: Tuple[Tuple, Tuple], e: A) -> Tuple[Tuple, Tuple]:
            l, r = z
            return (l + (e,), r) if f(e) else (l, r + (e,))
        l, r = reduce(splitter, self, ((), (),))  # type: ignore
        return List.wrap(l), List.wrap(r)

    def split_type(self, tpe: Union[Type, Tuple[Type[B]]]) -> Tuple['List[B]', 'List[A]']:
        return self.split(lambda a: isinstance(a, tpe))

    def index_of(self, target: A) -> 'maybe.Maybe[int]':
        return self.index_where(lambda a: a == target)

    @property
    def reversed(self) -> 'List[A]':
        return Lists.wrap(reversed(self))

    def mk_string(self, sep: str='') -> str:
        return sep.join(self / str)

    @property
    def join_lines(self) -> str:
        return self.mk_string('\n')

    @property
    def join_comma(self) -> str:
        return self.mk_string(', ')

    @property
    def join_tokens(self) -> str:
        return self.mk_string(' ')

    @property
    def join_dot(self) -> str:
        return self.mk_string('.')

    def cons(self, item: A) -> 'List[A]':
        return List.wrap(cons(item, self))

    def cons_m(self, item: 'maybe.Maybe[A]') -> 'List[A]':
        return item / self.cons | self

    def cat(self, item: A) -> 'List[A]':
        return self + List(item)

    def cat_m(self, item: 'maybe.Maybe[A]') -> 'List[A]':
        return item / self.cat | self

    @property
    def transpose(self) -> 'List[List[A]]':
        return List.wrap(map(List.wrap, zip(*self)))  # type: ignore

    def take(self, n: int) -> 'List[A]':
        return self[:n]

    def take_while(self, pred: Callable[[A], bool]) -> 'List[A]':
        index = self.index_where(lambda a: not pred(a))
        return index / (lambda a: self[:a]) | self

    def drop(self, n: int) -> 'List[A]':
        return self[n:]

    def drop_while(self, pred: Callable[[A], bool]) -> 'List[A]':
        index = self.index_where(lambda a: not pred(a))
        return index / (lambda a: self[a:]) | Nil

    def drop_while_or_self(self, pred: Callable[[A], bool]) -> 'List[A]':
        res = self.drop_while(pred)
        return self if res == Nil else res

    def drop_right(self, n: int) -> 'List[A]':
        return self.take(self.length - n)

    def remove_all(self, els: 'List[A]') -> 'List[A]':
        return self.filter_not(els.contains)

    def __repr__(self) -> str:
        return '{}({})'.format(self.__class__.__name__, ', '.join(map(repr, self)))

    def __str__(self) -> str:
        return '[{}]'.format(', '.join(map(safe_string, self)))

    def __hash__(self) -> int:
        return hash(tuple(self))

    def sort_by(self, f: Callable[[A], bool], reverse: bool=False) -> 'List[A]':
        return List.wrap(sorted(self, key=f, reverse=reverse))

    def sort(self, reverse: bool=False) -> 'List[A]':  # type: ignore
        return self.sort_by(I, reverse)

    def replace_item(self, a: A, b: A) -> 'List[A]':
        return self.map(lambda c: b if c == a else c)

    @curried
    def replace_where(self, a: A, pred: Callable) -> 'List[A]':
        return self.map(lambda b: a if pred(b) else b)

    def __mul__(self, count: int) -> 'List[A]':
        return List.wrap(super().__mul__(count))

    def group_by(self, f: Callable[[A], Any]) -> dict:
        from amino import Map
        return Map(groupby(f, self)).valmap(List.wrap)

    def slice(self, start: int, end: int) -> 'List[A]':
        return self[start:end]

    def indent(self, count: int) -> 'List[str]':
        ws = ' ' * count
        return self.map(lambda a: f'{ws}{a}')

    @property
    def rstrip(self) -> 'List[str]':
        return self / (lambda a: a.rstrip())

    @property
    def strip_newlines(self) -> 'List[str]':
        return self / (lambda a: a.replace('\n', ''))

    chomp = strip_newlines

    @property
    def rstrip_newlines(self) -> 'List[str]':
        return self / (lambda a: a.rstrip('\n'))

    def collect(self, f: Callable[[A], 'maybe.Maybe[B]']) -> 'List[B]':
        return self.flat_map(f)

    def modify_at(self, index: int, mod: Callable[[A], A]) -> 'maybe.Maybe[List[A]]':
        return self.lift(index).map(lambda a: self[:index].cat(mod(a)) + self[index + 1:])


class Lists:
    empty: List = List()

    @staticmethod
    def wrap(l: Iterable[B]) -> List[B]:
        return List(*list(l))

    @staticmethod
    def range(*a: int) -> List[int]:
        return List.wrap(range(*a))

    @staticmethod
    def random_string(num: int=10) -> str:
        chars = string.ascii_letters + string.digits
        return _rand_str(chars, num)

    @staticmethod
    def random_alpha(num: int=10) -> str:
        chars = string.ascii_letters
        return _rand_str(chars, num)

    @staticmethod
    def gen(num: int, f: Callable[[], List[A]]) -> List[A]:
        return List.range(num) // (lambda a: f())

    @staticmethod
    def fill(num: int, value: A) -> List[A]:
        return Lists.range(num).replace(value)

    @staticmethod
    def lines(data: str) -> List[str]:
        return List.wrap(data.splitlines())

    @staticmethod
    def split(data: str, splitter: str, maxsplit: int=-1) -> List[str]:
        return List.wrap(data.split(splitter, maxsplit))

    @staticmethod
    def rsplit(data: str, splitter: str, maxsplit: int=-1) -> List[str]:
        return List.wrap(data.rsplit(splitter, maxsplit))

    @staticmethod
    def tokens(data: str) -> List[str]:
        return Lists.split(data, ' ')

    @staticmethod
    @curried
    def iff(cond: bool, a: Union[A, Callable[[], A]]) -> List[A]:
        return List(call_by_name(a)) if cond else List()

    @staticmethod
    @curried
    def iff_l(cond: bool, a: Union[A, Callable[[], A]]) -> List[A]:
        return call_by_name(a) if cond else List()

    @staticmethod
    def file(path: Path) -> List[str]:
        return Lists.lines(path.read_text())

    @staticmethod
    def find_str_matches(data: str, target: str) -> List[int]:
        @tailrec
        def find(current: int, result: List[int]) -> TailrecResult[List[int]]:
            match = data.find(target, current)
            return (False, result) if match == -1 else (True, (match + 1, result.cat(match)))
        return find(0, Nil)


Nil = Lists.empty

__all__ = ('List', 'Lists', 'Nil')
